import unittest
from unittest.mock import patch, MagicMock
from neuronautics.ui.spike import Spike
from neuronautics.ui.abstract_source import AbstractSource
from neuronautics.recordings.nn_spike import NNSpike
from neuronautics.config.settings import Settings
import pandas as pd
import numpy as np

settings = Settings()


class TestSpike(unittest.TestCase):
    def test_constructor(self):
        mock_ui = MagicMock()
        spike = Spike(ui=mock_ui)
        self.assertIsInstance(spike, AbstractSource)
        self.assertIs(spike.ui, mock_ui)
        self.assertIs(spike.data_handler, NNSpike)

    def test_load_files(self):
        mock_ui = MagicMock()
        files = ['file1.spike', 'file2.spike']
        with patch('glob.glob') as spy_glob:
            spy_glob.return_value = files
            spike = Spike(mock_ui)
            spike.load_files('/random/folder')

            call_args_list = mock_ui.spikeBtn.setEnabled.call_args_list
            self.assertEqual(call_args_list[0].args, (False,))
            self.assertEqual(call_args_list[1].args, (True,))
            self.assertListEqual(spike.filenames, files)

    def test_load_files_empty_folder(self):
        mock_ui = MagicMock()
        files = []
        with patch('glob.glob') as spy_glob:
            spy_glob.return_value = files
            spike = Spike(mock_ui)
            spike.load_files('/random/folder')

            mock_ui.spikeBtn.setEnabled.assert_called_once_with(False)
            self.assertListEqual(spike.filenames, [])

    def test_select_view(self):
        with patch('neuronautics.ui.spike.AbstractSource.select_view') as spy_select_view:
            mock_ui = MagicMock()
            spike = Spike(mock_ui)

            spike.select_view()

            spy_select_view.assert_called_once()
            mock_ui.optionStck.setCurrentWidget.assert_called_once_with(mock_ui.spikeOptPage)

    def test_plot_single_view_stacked(self):
        mock_ui = MagicMock()
        mock_ui.chkSpikeStacked.isChecked.return_value = True
        mock_ui.recording_start_s = 0
        mock_ui.single_chart_view.x_label.return_value = mock_ui.single_chart_view
        mock_ui.single_chart_view.y_label.return_value = mock_ui.single_chart_view
        mock_ui.single_chart_view.title.return_value = mock_ui.single_chart_view
        mock_ui.single_chart_view.plot.return_value = mock_ui.single_chart_view

        spy_nnspike = MagicMock()
        spy_nnspike.return_value = spy_nnspike
        spy_nnspike.to_ms.side_effect = lambda x, _: x + 1
        spy_nnspike.get_channel_data.return_value = pd.DataFrame({'CLASS': [0]*3,
                                                                  'TS': [1, 2, 3],
                                                                  'SPK': [[1, 2], [3, 4], [4, 5]]})

        spike = Spike(mock_ui)
        spike.data_handler = spy_nnspike

        spike._plot_single_view(path='/random/file.spike', channel_id=0)

        mock_ui.single_chart_view.x_label.assert_called_once_with('Time (ms)')
        mock_ui.single_chart_view.y_label.assert_called_once_with('Microvolts')
        mock_ui.single_chart_view.title.assert_called_once_with('Channel 0')
        mock_ui.single_chart_view.plot.assert_called_once_with(
            values=[[(0, 1), (1, 2)],
                    [(0, 3), (1, 4)],
                    [(0, 4), (1, 5)]],
            group=[0, 0, 0],
            allow_selection=True,
            xlim=None,
            ylim=settings.ui_spike.MICRO_VOLTS_RANGE,
            progress_bar=True
        )

    def test_plot_single_view_timeseries(self):
        mock_ui = MagicMock()
        mock_ui.chkSpikeStacked.isChecked.return_value = False
        mock_ui.recording_start_s = 0
        mock_ui.single_chart_view.x_label.return_value = mock_ui.single_chart_view
        mock_ui.single_chart_view.y_label.return_value = mock_ui.single_chart_view
        mock_ui.single_chart_view.title.return_value = mock_ui.single_chart_view
        mock_ui.single_chart_view.plot.return_value = mock_ui.single_chart_view

        spy_nnspike = MagicMock()
        spy_nnspike.return_value = spy_nnspike
        spy_nnspike.to_ms.side_effect = lambda x, _: x + 1
        spy_nnspike.get_channel_data.return_value = pd.DataFrame({'CLASS': [0]*3,
                                                                  'TS': [1, 2, 3],
                                                                  'SPK': [[1, 2], [3, 4], [4, 5]]})

        spike = Spike(mock_ui)
        spike.data_handler = spy_nnspike

        spike._plot_single_view(path='/random/file.spike', channel_id=0)

        mock_ui.single_chart_view.x_label.assert_called_once_with('Time (ms)')
        mock_ui.single_chart_view.y_label.assert_called_once_with('Microvolts')
        mock_ui.single_chart_view.title.assert_called_once_with('Channel 0')
        mock_ui.single_chart_view.plot.assert_called_once_with(
            values=[[(1, 1), (2, 2)],
                    [(2, 3), (3, 4)],
                    [(3, 4), (4, 5)]],
            group=[0, 0, 0],
            allow_selection=True,
            xlim=(-1, settings.ui_spike.SINGLE_VIEW_LENGTH_S * 1_000),
            ylim=settings.ui_spike.MICRO_VOLTS_RANGE,
            progress_bar=True
        )

    def test_plot_grid_view_stacked(self):
        mock_ui = MagicMock()
        mock_ui.chkSpikeStacked.isChecked.return_value = True
        mock_ui.recording_start_s = 0

        spy_nnspike = MagicMock()
        spy_nnspike.return_value = spy_nnspike
        spy_nnspike.to_ms.side_effect = lambda x, _: x + 1
        spy_nnspike.get_all_data.return_value = pd.DataFrame({'CH_ID': [1, 2, 3],
                                                              'CLASS': [0]*3,
                                                              'TS': [1, 2, 3],
                                                              'SPK': [[1, 2], [3, 4], [4, 5]]})

        spike = Spike(mock_ui)
        spike.data_handler = spy_nnspike

        spike._plot_grid_view(path='/random/file.spike')

        expected_values = {
            1: ([[(0, 1), (1, 2)]], [0]),
            2: ([[(0, 3), (1, 4)]], [0]),
            3: ([[(0, 4), (1, 5)]], [0]),
        }
        mock_ui.multiple_chart_view.plot.assert_called_once_with(
            expected_values,
            xlim=None,
            ylim=settings.ui_spike.MICRO_VOLTS_RANGE
        )

    def test_plot_grid_view_timeseries(self):
        mock_ui = MagicMock()
        mock_ui.chkSpikeStacked.isChecked.return_value = False
        mock_ui.recording_start_s = 0

        spy_nnspike = MagicMock()
        spy_nnspike.return_value = spy_nnspike
        spy_nnspike.to_ms.side_effect = lambda x, _: x + 1
        spy_nnspike.get_all_data.return_value = pd.DataFrame({'CH_ID': [1, 2, 3],
                                                              'CLASS': [0]*3,
                                                              'TS': [1, 2, 3],
                                                              'SPK': [[1, 2], [3, 4], [4, 5]]})

        spike = Spike(mock_ui)
        spike.data_handler = spy_nnspike

        spike._plot_grid_view(path='/random/file.spike')

        expected_values = {
            1: ([[(1, 1), (2, 2)]], [0]),
            2: ([[(2, 3), (3, 4)]], [0]),
            3: ([[(3, 4), (4, 5)]], [0]),
        }
        mock_ui.multiple_chart_view.plot.assert_called_once_with(
            expected_values,
            xlim=(-1, settings.ui_spike.GRID_VIEW_TIMESERIES_S * 1_000),
            ylim=settings.ui_spike.MICRO_VOLTS_RANGE
        )

    def test_plot_list_view_stacked(self):
        mock_ui = MagicMock()
        mock_ui.chkSpikeStacked.isChecked.return_value = True
        mock_ui.recording_start_s = 0

        spy_nnspike = MagicMock()
        spy_nnspike.return_value = spy_nnspike
        spy_nnspike.to_ms.side_effect = lambda x, _: x + 1
        spy_nnspike.get_channel_data.return_value = pd.DataFrame({'CLASS': [0]*3,
                                                                  'TS': [1, 2, 3],
                                                                  'SPK': [[1, 2], [3, 4], [4, 5]]})

        spike = Spike(mock_ui)
        spike.filenames = ['file.spike']
        spike.data_handler = spy_nnspike

        spike._plot_list_view(channel_id=0)

        expected_values = {
            'file': ([[(0, 1), (1, 2)],
                      [(0, 3), (1, 4)],
                      [(0, 4), (1, 5)]], [0, 0, 0]),
        }
        mock_ui.list_chart_view.plot.assert_called_once_with(
            expected_values,
            xlim=None,
            ylim=settings.ui_spike.MICRO_VOLTS_RANGE
        )

    def test_plot_list_view_timeseries(self):
        mock_ui = MagicMock()
        mock_ui.chkSpikeStacked.isChecked.return_value = False
        mock_ui.recording_start_s = 0

        spy_nnspike = MagicMock()
        spy_nnspike.return_value = spy_nnspike
        spy_nnspike.to_ms.side_effect = lambda x, _: x + 1
        spy_nnspike.get_channel_data.return_value = pd.DataFrame({'CLASS': [0]*3,
                                                                  'TS': [1, 2, 3],
                                                                  'SPK': [[1, 2], [3, 4], [4, 5]]})

        spike = Spike(mock_ui)
        spike.filenames = ['file.spike']
        spike.data_handler = spy_nnspike

        spike._plot_list_view(channel_id=0)

        expected_values = {
            'file': ([[(1, 1), (2, 2)],
                      [(2, 3), (3, 4)],
                      [(3, 4), (4, 5)]], [0, 0, 0]),
        }
        mock_ui.list_chart_view.plot.assert_called_once_with(
            expected_values,
            xlim=(-1, settings.ui_spike.GRID_VIEW_TIMESERIES_S*1_000),
            ylim=settings.ui_spike.MICRO_VOLTS_RANGE
        )

    def test_all_data(self):
        mock_ui = MagicMock()
        mock_ui.fileSelector.currentText.return_value = 'file.spike'

        spy_nnspike = MagicMock()
        spy_nnspike.return_value = spy_nnspike
        expected_df = pd.DataFrame({'CLASS': [0]*3,
                                    'TS': [1, 2, 3],
                                    'SPK': [[1, 2], [3, 4], [4, 5]]})
        spy_nnspike.get_all_data.return_value = expected_df

        spike = Spike(mock_ui)
        spike.data_handler = spy_nnspike

        df = spike.all_data()

        spy_nnspike.assert_called_once_with('file.spike')
        spy_nnspike.get_all_data.assert_called_once_with(0, np.Inf)
        pd.testing.assert_frame_equal(df, expected_df)

    def test_all_channel_data(self):
        mock_ui = MagicMock()
        mock_ui.fileSelector.currentText.return_value = 'file.spike'
        mock_ui.channelSelector.currentText.return_value = '1'

        spy_nnspike = MagicMock()
        spy_nnspike.return_value = spy_nnspike
        expected_df = pd.DataFrame({'CLASS': [0]*3,
                                    'TS': [1, 2, 3],
                                    'SPK': [[1, 2], [3, 4], [4, 5]]})
        spy_nnspike.get_channel_data.return_value = expected_df

        spike = Spike(mock_ui)
        spike.data_handler = spy_nnspike

        df = spike.all_channel_data()

        spy_nnspike.assert_called_once_with('file.spike')
        spy_nnspike.get_channel_data.assert_called_once_with(1, 0, np.Inf)
        pd.testing.assert_frame_equal(df, expected_df)

    def test_run_kmeans(self):
        mock_ui = MagicMock()

        spy_nnspike = MagicMock()
        spy_nnspike.return_value = spy_nnspike

        spike = Spike(mock_ui)
        spike.data_handler = spy_nnspike
        spike._get_kmeans_params = MagicMock()
        spike._get_kmeans_params.return_value = ('file.spike', 0, 2, 10)
        spike._stimulus_params = MagicMock()
        spike._stimulus_params.return_value = (1, 2, 3)
        spike.all_channel_data = MagicMock()
        df = pd.DataFrame({'class': [1, 2, 3]})
        spike.all_channel_data.return_value = df
        spike._pca_kmeans = MagicMock()
        spike._relabel = MagicMock()
        spike._relabel.return_value = [1, -1, 0]
        spike.plot_single_view = MagicMock()

        spike.run_kmeans()

        spike._pca_kmeans.assert_called_once_with(df, 2, 10)
        spike.plot_single_view.assert_called_once()
        spy_nnspike.set_labels.assert_called_once_with(0, [1, -1, 0])

    def test_stimulus_params(self):
        mock_ui = MagicMock()
        mock_ui.stimUp.value.return_value = 1
        mock_ui.stimDown.value.return_value = 2
        mock_ui.stimShift.value.return_value = 3

        spike = Spike(mock_ui)

        stim_up, stim_down, stim_shift = spike._stimulus_params()
        self.assertEqual(stim_up, 1)
        self.assertEqual(stim_down, 2)
        self.assertEqual(stim_shift, 3)

    def test_get_kmeans_params(self):
        mock_ui = MagicMock()
        mock_ui.fileSelector.currentText.return_value = 'file.spike'
        mock_ui.channelSelector.currentText.return_value = '123'
        mock_ui.pcaValue.value.return_value = 4
        mock_ui.kValue.value.return_value = 10

        spike = Spike(mock_ui)

        (path, channel_id, num_pca, num_clusters) = spike._get_kmeans_params()

        self.assertEqual(path, 'file.spike')
        self.assertEqual(channel_id, 123)
        self.assertEqual(num_pca, 4)
        self.assertEqual(num_clusters, 10)

    def test_pca_kmeans(self):
        mock_ui = MagicMock()
        arr = [[1, 2, 3], [3, 4, 5]]
        input_df = pd.DataFrame({'spike': arr})

        with patch('neuronautics.ui.spike.SpikeSorter') as spy_spike_sorter:
            spike = Spike(mock_ui)
            labels = spike._pca_kmeans(input_df, 2, 4)

            np.testing.assert_array_equal(spy_spike_sorter.call_args.args[0], arr)
            spy_spike_sorter.return_value.detect_noise.assert_called_once()
            spy_spike_sorter.return_value.detect_noise.return_value.pca.assert_called_once_with(2)
            spy_spike_sorter.return_value.detect_noise.return_value.pca.return_value.kmeans.assert_called_once_with(4)
            spy_spike_sorter.return_value.detect_noise.return_value.pca.return_value.kmeans.return_value.run.assert_called_once()

            self.assertEqual(labels,
                             spy_spike_sorter.return_value
                                             .detect_noise.return_value
                                             .pca.return_value
                                             .kmeans.return_value
                                             .run.return_value)

    def test_relabel(self):
        mock_ui = MagicMock()

        spike = Spike(mock_ui)

        df = pd.DataFrame({'ts_ms': [1,2,3,4,5,6]})
        labels = [0,0,0,0,0,0]
        stim_up, stim_down, stim_shift = 0.5, 1.5, 1

        new_labels = spike._relabel(df, labels, stim_up, stim_down, stim_shift)
        self.assertListEqual(new_labels, [-1, 0, -1, 0, -1, 0])

    def test_relabel_keeping_noise(self):
        mock_ui = MagicMock()

        spike = Spike(mock_ui)

        df = pd.DataFrame({'ts_ms': [1,2,3,4,5,6]})
        labels = [-1] * 6
        stim_up, stim_down, stim_shift = 0.5, 1.5, 1

        new_labels = spike._relabel(df, labels, stim_up, stim_down, stim_shift)
        self.assertListEqual(new_labels, [-1, -1, -1, -1, -1, -1])

    def test_relabel_keep_labels(self):
        mock_ui = MagicMock()

        spike = Spike(mock_ui)

        df = pd.DataFrame({'ts_ms': [1,2,3,4,5,6]})
        labels = [-1] * 9
        stim_up, stim_down, stim_shift = 0.5, 1.5, 1

        new_labels = spike._relabel(df, labels, stim_up, stim_down, stim_shift)
        self.assertListEqual(new_labels, [-1, -1, -1, -1, -1, -1, -1, -1, -1])


    def test_run_all_kmeans(self):
        mock_ui = MagicMock()

        spy_nnspike = MagicMock()
        spy_nnspike.return_value = spy_nnspike
        input_df = pd.DataFrame({'spike': [[1, 2, 3], [3, 4, 5], [11, 12, 13], [13, 14, 15]],
                                 'channel_id': [0, 0, 1, 1],
                                 'class': [0, 0, 0, 0]})
        spy_nnspike.get_all_data.return_value = input_df

        spike = Spike(mock_ui)
        spike.data_handler = spy_nnspike
        spike._stimulus_params = MagicMock()
        spike._stimulus_params.return_value = (1, 2, 3)
        spike._get_kmeans_params = MagicMock()
        spike._get_kmeans_params.return_value = ('file.spike', 0, 2, 10)
        spike._pca_kmeans = MagicMock()
        spike.plot_single_view = MagicMock()
        spike.plot_multiple_view = MagicMock()
        spike._relabel = MagicMock()
        spike._relabel.return_value = [1, -1, 0, 2]

        spike.run_all_kmeans()

        called_with_channel_0 = pd.DataFrame({'spike': [[1, 2, 3], [3, 4, 5]],
                                              'channel_id': [0, 0],
                                              'class': [0, 0]})

        called_with_channel_1 = pd.DataFrame({'spike': [[11, 12, 13], [13, 14, 15]],
                                              'channel_id': [1, 1],
                                              'class': [0, 0]})

        args = spike._pca_kmeans.call_args_list
        pd.testing.assert_frame_equal(args[0].args[0], called_with_channel_0)
        np.testing.assert_array_equal(args[1].args[0].values, called_with_channel_1.values )
        spike.plot_single_view.assert_called_once()
        spike.plot_multiple_view.assert_called_once()
        spy_nnspike.set_labels.assert_called_with(1, [1, -1])

    def test_reset_spike_classes(self):
        mock_ui = MagicMock()

        spy_nnspike = MagicMock()
        spy_nnspike.return_value = spy_nnspike
        input_df = pd.DataFrame({'spike': [[1, 2, 3], [3, 4, 5], [11, 12, 13]]})
        spy_nnspike.get_channel_data.return_value = input_df

        spike = Spike(mock_ui)
        spike.data_handler = spy_nnspike
        spike._get_kmeans_params = MagicMock()
        spike._get_kmeans_params.return_value = ('file.spike', 0, 2, 10)
        spike.plot_single_view = MagicMock()

        spike.reset_spike_classes()

        spy_nnspike.set_labels.assert_called_once_with(0, [0, 0, 0])
        spike.plot_single_view.assert_called_once()

    def test_signal_unit_changed(self):
        mock_ui = MagicMock()

        spy_nnspike = MagicMock()
        spy_nnspike.return_value = spy_nnspike
        input_df = pd.DataFrame({'spike': [[1, 2, 3], [3, 4, 5], [11, 12, 13]],
                                 'ts_ms': [100, 200, 300],
                                 'class': [0, 1, 2]})
        spy_nnspike.get_channel_data.return_value = input_df

        spike = Spike(mock_ui)
        spike.data_handler = spy_nnspike
        spike._get_kmeans_params = MagicMock()
        spike._get_kmeans_params.return_value = ('file.spike', 0, 2, 10)

        spike.signal_unit_changed([(0, 2), (2, -1)])  # index, class

        spy_nnspike.set_labels.assert_called_once_with(0, [2, 1, -1])

    def test_save_spikes(self):
        mock_ui = MagicMock()
        mock_ui.fileSelector.currentText.return_value='file.spike'

        spy_nnspike = MagicMock()
        spy_nnspike.return_value = spy_nnspike

        spike = Spike(mock_ui)
        spike.data_handler = spy_nnspike

        spike.save_spikes()

        spy_nnspike.assert_called_once_with('file.spike')
        spy_nnspike.save.assert_called_once()


if __name__ == '__main__':
    unittest.main()




import os
import pkg_resources

MAX_WORKERS = os.cpu_count() - 1


UI_NEURONAUTICS = pkg_resources.resource_filename("neuronautics.ui", "neuronautics.ui")
UI_LAYOUT_CREATION = pkg_resources.resource_filename("neuronautics.ui", "layout-creation.ui")
UI_ANALYSIS_CREATION = pkg_resources.resource_filename("neuronautics.ui", "analysis-creation.ui")
UI_CODE_ERROR = pkg_resources.resource_filename("neuronautics.ui", "code-error.ui")

JINJA_GRAPH_ANALYSIS = pkg_resources.resource_filename("neuronautics.analysis.templates", "graph-analysis.jinja")
JINJA_BAR_ANALYSIS = pkg_resources.resource_filename("neuronautics.analysis.templates", "bar-analysis.jinja")
JINJA_LINE_ANALYSIS = pkg_resources.resource_filename("neuronautics.analysis.templates", "line-analysis.jinja")
JINJA_IMAGE_ANALYSIS = pkg_resources.resource_filename("neuronautics.analysis.templates", "image-analysis.jinja")
JINJA_BASE_ANALYSIS = pkg_resources.resource_filename("neuronautics.analysis.templates", "base-analysis.jinja")

YML_ANALYSIS_BASE = pkg_resources.resource_filename("neuronautics.analysis", "analysis-base.yml")


from neuronautics.analysis.type.graph_analysis import GraphAnalysis
from neuronautics.analysis.helpers import to_timeseries
from neuronautics.config.layout import Layout
import numpy as np
import neo
import quantities as pq
from elephant.spike_train_correlation import spike_time_tiling_coefficient


class SpikeTimeTilingCoefficient(GraphAnalysis):

    def get_input_params(self):
        """Get the input parameters for the activity correlation analysis.

        Returns:
            list: A list of dictionaries, each describing an input parameter.
                Each dictionary includes 'name', 'min', 'max', 'default', and 'type' keys.

        """
        return [
            {'name': 'corr_thr', 'min': 0, 'max': 1, 'default': 0.3, 'type': 'float'}
        ]

    def run(self, spikes, corr_thr, *args, **kwargs):
        """Run the activity correlation analysis.

        Args:
            spikes (DataFrame): A DataFrame containing spike data.
            corr_thr (float): The correlation threshold for considering correlations.

        Returns:
            ndarray: A boolean matrix indicating correlations above the threshold.

        """
        spikes['class'] = spikes['class'].astype(int)
        spikes = spikes[spikes['class'] >= 0]  # remove noise  TODO: hardcoded 0
        spikes = spikes[['channel_id', 'ts_ms']].copy().reset_index(drop=True)
        max_ms = spikes.ts_ms.max()
        spikes = spikes.groupby('channel_id').agg(list)
        spikes = spikes.to_dict()['ts_ms']

        spikes = {ch_id: neo.SpikeTrain(spike, units='ms', t_stop=max_ms) for (ch_id, spike) in spikes.items()}

        layout = Layout().current()
        max_lay = np.max([np.max([int(l) for lay in layout for l in lay if l != ''])])
        w_ij = np.zeros((max_lay, max_lay))

        channel_ids = list(spikes.keys())

        for ix, ch1 in enumerate(channel_ids):
            for ch2 in channel_ids[ix+1:]:
                coef = spike_time_tiling_coefficient(spikes[ch1], spikes[ch2])
                w_ij[ch1, ch2] = coef
                w_ij[ch2, ch1] = coef

        return w_ij > corr_thr
    
    def plot(self, *args, **kwargs):
        """Plot the activity correlation analysis.

        Args:
            *args: Variable-length positional arguments.
            **kwargs: Variable-length keyword arguments.

        Returns:
            object: matplotlib figure

        """
        return super().plot('Spike Time Tiling Coefficient', *args, **kwargs)



import unittest
from unittest.mock import patch, MagicMock
from neuronautics.analysis.type.image_analysis import ImageAnalysis
from neuronautics.analysis.type.abstract_analysis import AnalysisType
from matplotlib.figure import Figure
import pandas as pd
import numpy as np


class OneImageAnalysis(ImageAnalysis):

    def run(self, *args, **kwargs):
        return np.array([[0, 1], [1, 0]])

    def get_input_params(self):
        pass


class TestImageAnalysis(unittest.TestCase):
    def test_type(self):
        analysis = OneImageAnalysis()
        self.assertEqual(analysis.type(), AnalysisType.IMAGE)

    @patch("matplotlib.pyplot.subplots")
    def test_plot(self, mock_subplots):
        mock_fig = MagicMock()
        mock_ax = MagicMock()
        mock_subplots.return_value = (mock_fig, mock_ax)
        # Act
        analysis = OneImageAnalysis()
        result = analysis.plot(title='my plot')

        # Assert
        self.assertEqual(result, mock_fig)
        mock_ax.imshow.assert_called_once()

    @patch("matplotlib.pyplot.colorbar")
    @patch("matplotlib.pyplot.subplots")
    def test_plot_cmap(self, mock_subplots, mock_colorbar):
        mock_fig = MagicMock()
        mock_ax = MagicMock()
        mock_subplots.return_value = (mock_fig, mock_ax)
        # Act
        analysis = OneImageAnalysis()
        result = analysis.plot(title='my plot', cmap='Reds')

        # Assert
        self.assertEqual(result, mock_fig)
        mock_ax.imshow.assert_called_once()
        mock_colorbar.assert_called_once()


if __name__ == '__main__':
    unittest.main()



from abc import ABC

from neuronautics.analysis.type.abstract_analysis import AbstractAnalysis, AnalysisType
from neuronautics.analysis.type.palette import Palette
import matplotlib.pyplot as plt


class LineAnalysis(AbstractAnalysis, ABC):

    def type(self):
        return AnalysisType.LINE

    def plot(self, xlabel='', ylabel='', title='', *args, **kwargs):
        dataframe = self.run(*args, **kwargs)

        x_values = dataframe.iloc[:, 0]
        y_values = dataframe.iloc[:, 1:]

        fig, ax = plt.subplots(figsize=(10, 6))

        for col in y_values.columns:
            ax.plot(x_values, y_values[col], label=col)

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(title)
        plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))  # Place legend outside the plot on the right
        plt.grid(True)

        return fig



import yaml
from pathlib import Path
from enum import Enum

from ..utils.helpers import file_path
from ..utils.singleton import Singleton

# TODO: tiene que ser capaz de leer y modificar en tiempo real


class SubSettings:
    def __getattr__(self, item):
        pass


class Settings(metaclass=Singleton):
    FILE = file_path('settings.yml')

    def __init__(self):
        settings = self.load()
        self.ui_spike = UiSpikeSettings(**settings)
        self.ui_raw = UiRawSettings(**settings)

    def save(self):
        settings = {k: v.__dict__ for k,v in self.__dict__.items()}
        with open(Settings.FILE, 'w') as stream:
            yaml.dump(settings, stream)

    @classmethod
    def load(cls):
        if Path(Settings.FILE).exists():
            with open(Settings.FILE, 'r') as stream:
                data = yaml.full_load(stream)
                if not data:
                    return dict()
                return data
        else:
            return dict()


class UiSpikeSettings(SubSettings, metaclass=Singleton):

    def __init__(self, ui_spike=None, **kwargs):
        self.SINGLE_VIEW_LENGTH_S = 5
        self.GRID_VIEW_STACKED_S = 30
        self.GRID_VIEW_TIMESERIES_S = 3
        self.MICRO_VOLTS_RANGE = None  # None or tuple e.g (-200, 200)
        if ui_spike:
            for key in ui_spike:
                setattr(self, key, ui_spike[key])

    def set_single_view(self, val):
        self.SINGLE_VIEW_LENGTH_S = val

    def set_grid_stacked_view(self, val):
        self.GRID_VIEW_STACKED_S = val

    def set_grid_timeseries_view(self, val):
        self.GRID_VIEW_TIMESERIES_S = val

    def set_volts_range(self, min_val, max_val):
        if min_val == 0 and max_val == 0:
            self.MICRO_VOLTS_RANGE = None
        else:
            self.MICRO_VOLTS_RANGE = (min_val, max_val)

    def set_volts_min(self, min_val):
        if self.MICRO_VOLTS_RANGE is None:
            self.set_volts_range(min_val, 0)
        else:
            self.set_volts_range(min_val, self.MICRO_VOLTS_RANGE[1])

    def set_volts_max(self, max_val):
        if self.MICRO_VOLTS_RANGE is None:
            self.set_volts_range(0, max_val)
        else:
            self.set_volts_range(self.MICRO_VOLTS_RANGE[0], max_val)


class UiRawSettings(SubSettings, metaclass=Singleton):

    def __init__(self, ui_raw=None, **kwargs):
        self.SINGLE_VIEW_LENGTH_S = 5
        self.GRID_VIEW_LENGTH_S = .3
        self.MICRO_VOLTS_RANGE = None
        if ui_raw:
            for key in ui_raw:
                setattr(self, key, ui_raw[key])

    def set_single_view(self, val):
        self.SINGLE_VIEW_LENGTH_S = val

    def set_grid_view(self, val):
        self.GRID_VIEW_LENGTH_S = val

    def set_volts_range(self, min_val, max_val):
        if min_val == 0 and max_val == 0:
            self.MICRO_VOLTS_RANGE = None
        else:
            self.MICRO_VOLTS_RANGE = (min_val, max_val)

    def set_volts_min(self, min_val):
        if self.MICRO_VOLTS_RANGE is None:
            self.set_volts_range(min_val, 0)
        else:
            self.set_volts_range(min_val, self.MICRO_VOLTS_RANGE[1])

    def set_volts_max(self, max_val):
        if self.MICRO_VOLTS_RANGE is None:
            self.set_volts_range(0, max_val)
        else:
            self.set_volts_range(self.MICRO_VOLTS_RANGE[0], max_val)



from abc import ABC

from neuronautics.analysis.type.abstract_analysis import AbstractAnalysis, AnalysisType


class MiscAnalysis(AbstractAnalysis, ABC):

    def type(self):
        return AnalysisType.OTHER



from PyQt5 import QtWidgets
import abc


class AbstractSource(abc.ABC):
    def __init__(self, ui, data_handler):
        self.ui = ui
        self.filenames = []
        self.data_handler = data_handler

    @abc.abstractmethod
    def load_files(self, folder_path):
        raise NotImplementedError("Not implemented")

    def select_view(self):
        self.ui.fileSelector.clear()
        for file in self.filenames:
            self.ui.fileSelector.addItem(file)
        self.select_file()

    def select_file(self):
        self.load_analog_channels_ids()

        self.plot_info_view()
        self.plot_multiple_view()
        self.plot_list_view()

    def load_analog_channels_ids(self):
        path = self.ui.fileSelector.currentText()
        if path is None or len(path) == 0:
            return

        data_source = self.data_handler(path)
        channel_ids = data_source.get_analog_channels_ids()
        self.ui.channelSelector.clear()
        self.ui.channelSelector2.clear()
        for ch_id in channel_ids:
            self.ui.channelSelector.addItem(str(ch_id))
            self.ui.channelSelector2.addItem(str(ch_id))
        self.ui.recording_start_s = 0

    def plot_info_view(self):
        path = self.ui.fileSelector.currentText()
        if not path:
            return
        data_source = self.data_handler(path)
        info = data_source.get_info()

        row = 0
        self.ui.infoTable.setRowCount(len(info))
        for key, value in info.items():
            key_item = QtWidgets.QTableWidgetItem(key)
            value_item = QtWidgets.QTableWidgetItem(value)

            self.ui.infoTable.setItem(row, 0, key_item)
            self.ui.infoTable.setItem(row, 1, value_item)

            row += 1
        self.ui.infoTable.setHorizontalHeaderLabels(["Key", "Value"])
        self.ui.infoTable.resizeColumnsToContents()

    def plot_single_view(self):
        path = self.ui.fileSelector.currentText()
        channel_id = self.ui.channelSelector.currentText()
        if self.ui.is_single_view_selected() and path and channel_id:
            self._plot_single_view(path, int(channel_id))
            self.ui.update()

    @abc.abstractmethod
    def _plot_single_view(self, path, channel_id):
        raise NotImplementedError('Not implemented')

    def plot_multiple_view(self):
        path = self.ui.fileSelector.currentText()
        if self.ui.is_grid_view_selected() and path:
            self._plot_grid_view(path)
            self.ui.update()

    @abc.abstractmethod
    def _plot_grid_view(self, path):
        raise NotImplementedError('Not implemented')

    def plot_list_view(self):
        channel_id = self.ui.channelSelector.currentText()
        if self.ui.is_list_view_selected() and channel_id:
            self._plot_list_view(int(channel_id))
            self.ui.update()

    @abc.abstractmethod
    def _plot_list_view(self, channel_id):
        raise NotImplementedError('Not implemented')




from datetime import datetime
from PyQt5.QtCore import QObject, pyqtSignal


class Logger(QObject):
    """
    Provides logging capabilities for tracking and reporting process progress.

    This class allows you to log the progress of processes, estimating remaining
    time, and emitting signals for progress updates.

    Attributes:
        process_time (dict): A dictionary storing process titles as keys and
                            corresponding start and end time instances as values.
        progress_signal (pyqtSignal): A signal emitted on progress updates, providing
                                     a formatted message and progress percentage.
        me (Logger): A class-level instance of the Logger class, acting as a singleton.

    Methods:
        get_logger(cls): Returns the singleton instance of the Logger class.
        log_process(self, title, current_step, total_steps): Logs process progress,
                                                             estimating remaining time,
                                                             and emits progress signals.
    """

    process_time = dict()
    progress_signal = pyqtSignal(str, int)
    me = None

    @classmethod
    def get_logger(cls):
        """
        Returns a singleton instance of the Logger class.

        This class method returns an instance of the Logger class if it doesn't
        exist already. The Logger class utilizes a singleton pattern, ensuring
        that only one instance of the class is created and shared across calls.

        Returns:
            Logger: A singleton instance of the Logger class.
        """
        if cls.me is None:
            cls.me = Logger()
        return cls.me

    def log_process(self, title, current_step, total_steps):
        """
        Logs the progress of a process, estimates remaining time, and emits
        progress signals.

        Args:
            title (str): The title of the process.
            current_step (int): The current step of the process.
            total_steps (int): The total number of steps in the process.
        """
        percent = 0 if total_steps == 0 else current_step / total_steps
        times = self.process_time.get(title, [])
        times.append(datetime.now())
        self.process_time[title] = times
        msg = '??'
        if len(times) > 1:
            t0 = times[0]
            delta = 0
            for t in times[1:]:
                delta += (t-t0).total_seconds()
                t0 = t
            avg_delta = delta / (len(times) - 1)
            remaining_time = avg_delta * (total_steps - current_step)
            hours, remainder = divmod(remaining_time, 3600)
            minutes, secs = divmod(remainder, 60)
            msg = f'{hours:02.0f}:{minutes:02.0f}:{secs:05.2f}'

        if current_step >= total_steps:
            total_time = (times[-1]-times[0]).total_seconds()

            hours, remainder = divmod(total_time, 3600)
            minutes, secs = divmod(remainder, 60)
            msg = f'{hours:02.0f}:{minutes:02.0f}:{secs:05.2f}'
            msg = f'Total time [{msg}]'
            self.process_time.pop(title)

        self.progress_signal.emit(msg, int(100*percent))



# logger
from ..utils.logger import Logger

# pyqt
from PyQt5 import QtWidgets, uic
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QTextCharFormat, QColor, QTextCursor

from ..recordings.config import UI_CODE_ERROR

import traceback
import re

REMAIN_COLOR = '#333333'
PATH_COLOR = '#1f77b4'
CODE_COLOR = '#33a02c'
EXCEPTION_COLOR = '#e31a1c'


class CodeErrorUi(QtWidgets.QDialog):
    def __init__(self):
        super(CodeErrorUi, self).__init__()
        uic.loadUi(UI_CODE_ERROR, self)

        self.errorText.clear()
        self.errorText.setTextColor(QColor(REMAIN_COLOR))
        self.highlight_traceback(str(traceback.format_exc()))

        self.showMaximized()

    def highlight_traceback(self, traceback_msg):
        path_pattern = re.compile(r'File ".*?"')
        code_pattern = re.compile(r'(?<=line )\d+')
        exception_pattern = re.compile(r'\b\w+Error\b')

        self.errorText.append(traceback_msg)

        path_format = QTextCharFormat()
        path_format.setForeground(QColor(PATH_COLOR))

        code_format = QTextCharFormat()
        code_format.setForeground(QColor(CODE_COLOR))

        exception_format = QTextCharFormat()
        exception_format.setForeground(QColor(EXCEPTION_COLOR))

        cursor = self.errorText.textCursor()

        for match in path_pattern.finditer(traceback_msg):
            cursor.setPosition(match.start())
            cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, len(match.group()))
            cursor.setCharFormat(path_format)

        for match in code_pattern.finditer(traceback_msg):
            cursor.setPosition(match.start())
            cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, len(match.group()))
            cursor.setCharFormat(code_format)

        for match in exception_pattern.finditer(traceback_msg):
            cursor.setPosition(match.start())
            cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, len(match.group()))
            cursor.setCharFormat(exception_format)

    @classmethod
    def show_exception(cls):
        ui = CodeErrorUi()
        ui.exec_()




import unittest
from unittest.mock import patch, MagicMock
from neuronautics.utils.logger import Logger
from datetime import datetime, timedelta


class TestLogger(unittest.TestCase):
    def test_get_logger_singleton(self):
        instance1 = Logger.get_logger()
        instance2 = Logger.get_logger()
        self.assertEqual(instance1, instance2)  # add assertion here

    def test_log_process_first_log(self):
        logger = Logger.get_logger()
        logger.process_time.clear() # ensuring first log
        logger.progress_signal = MagicMock()

        logger.log_process('title', 3, 100)
        logger.progress_signal.emit.assert_called_once_with('??', 3)

    def test_log_process(self):
        logger = Logger.get_logger()
        logger.process_time.clear() # ensuring first log
        logger.progress_signal = MagicMock()

        logger.log_process('title', 1, 100)
        logger.log_process('title', 10, 100)

        calls = logger.progress_signal.emit.call_args_list

        self.assertEqual(calls[0].args, ('??', 1))
        self.assertEqual(calls[1].args[1], 10)
        self.assertRegex(calls[1].args[0], r'00:0\d:\d\d.\d\d')

    def test_log_process_end_msg(self):
        logger = Logger.get_logger()
        logger.process_time.clear() # ensuring first log
        logger.progress_signal = MagicMock()

        logger.log_process('title', 1, 100)
        logger.log_process('title', 100, 100)

        calls = logger.progress_signal.emit.call_args_list

        self.assertEqual(calls[0].args, ('??', 1))
        self.assertEqual(calls[1].args[1], 100)
        self.assertRegex(calls[1].args[0], r'Total time \[00:0\d:\d\d.\d\d\]')


if __name__ == '__main__':
    unittest.main()



# data handlers
from ..ui.raw import Raw
from ..ui.spike import Spike

from ..analysis.loader import Loader

# plotters
from ..plotters.line_2d import Line2D
from ..plotters.grid_line_2d import GridLine2D

# ui handlers
from ..ui.analysis_ui import AnalysisUi
from ..ui.layout_ui import LayoutUi
from ..ui.settings_ui import SettingsUi
from ..recordings.config import UI_NEURONAUTICS

# logger
from ..utils.logger import Logger

# pyqt
from PyQt5 import QtWidgets, uic
from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import QMessageBox

CLOCK_SPEED_MS = 10
RECORDING_STEP_S = 0.1


class NeuronauticsUi(QtWidgets.QMainWindow):
    def __init__(self):
        super(NeuronauticsUi, self).__init__()
        uic.loadUi(UI_NEURONAUTICS, self)
        self.single_chart_view = Line2D(self.singleChart)
        self.single_chart_view.signal_unit_changed.connect(self.signal_unit_changed)

        self.multiple_chart_view = GridLine2D(self.multipleChart).setup_grid(nrows=8, ncols=15)
        self.list_chart_view = GridLine2D(self.listChart)

        self.recording_start_s, self.recording_step_s = 0, RECORDING_STEP_S
        self.update_timer = QTimer()
        self.update_timer.timeout.connect(self.update_canvas)

        self.logger = Logger.get_logger()
        self.logger.progress_signal.connect(self.update_progress)

        self.ui_view = None
        self.raw_handler = Raw(self)
        self.spike_handler = Spike(self)

        self.layout_ui = LayoutUi(self)
        self.layoutTbl.itemChanged.connect(self.layout_ui.handle_table_item_change)
        self.load_layout_names()

        self.loader = Loader()
        self.analysis_ui = AnalysisUi(self)
        self.settings_ui = SettingsUi(self)

        self.load_analysis()

        self.showMaximized()

    def closeEvent(self, event):
        # Display a confirmation message box
        confirm_dialog = QMessageBox(self)
        confirm_dialog.setIcon(QMessageBox.Question)
        confirm_dialog.setWindowTitle('Confirm Close')
        confirm_dialog.setText('Are you sure you want to close the application?')
        confirm_dialog.setStandardButtons(QMessageBox.Yes | QMessageBox.No)

        result = confirm_dialog.exec_()

        if result == QMessageBox.Yes:
            event.accept()  # Close the application
        else:
            event.ignore()  # Continue running the application

    def load_file(self):
        if self.ui_view is not None:
            self.ui_view.select_file()

    def update_progress(self, msg, progress):
        self.progressBar.setValue(progress)
        self.progressBar.setTextVisible(True)
        self.progressBar.setFormat(msg)

    def timer_start(self):
        self.update_timer.start(CLOCK_SPEED_MS)

    def timer_stop(self):
        if self.update_timer.isActive():
            self.update_timer.stop()
        else:
            self.recording_start_s = 0
            self.plot_view()

    def step_forward(self):
        if self.ui_view is not None:
            self.recording_start_s += self.recording_step_s * self.speedSelector.value()
            self.plot_view()

    def step_backward(self):
        if self.ui_view is not None:
            self.recording_start_s -= self.recording_step_s * self.speedSelector.value()
            self.recording_start_s = max(self.recording_start_s, 0)
            self.plot_view()

    def update_canvas(self):
        if self.ui_view is not None:
            self.step_forward()

    def plot_view(self):

        if self.ui_view is not None:
            if self.is_info_view_selected():
                self.ui_view.plot_info_view()
            elif self.is_single_view_selected():
                self.ui_view.plot_single_view()
            elif self.is_grid_view_selected():
                self.ui_view.plot_multiple_view()
            elif self.is_list_view_selected():
                self.ui_view.plot_list_view()

    def is_info_view_selected(self):
        return self.stackedWidget.currentWidget() == self.infoView

    def is_single_view_selected(self):
        return self.stackedWidget.currentWidget() == self.singleView

    def is_grid_view_selected(self):
        return self.stackedWidget.currentWidget() == self.gridView

    def is_list_view_selected(self):
        return self.stackedWidget.currentWidget() == self.listView

    def open_folder(self):
        folder_path = QtWidgets.QFileDialog.getExistingDirectory(self, "Select Folder", options=QtWidgets.QFileDialog.DontUseNativeDialog)
        if folder_path:
            self.load_files(folder_path)

    def load_files(self, folder_path):
        self.fileSelector.clear()
        self.raw_handler.load_files(folder_path)
        self.spike_handler.load_files(folder_path)
        self.analysisBtn.setEnabled(True)

        self.ui_view = None

    def raw_view(self):
        self.ui_view = self.raw_handler
        self.ui_view.select_view()

    def spike_view(self):
        self.ui_view = self.spike_handler
        self.ui_view.select_view()

    def select_info_view(self):
        self.stackedWidget.setCurrentWidget(self.infoView)
        self.optionStck.setCurrentWidget(self.emptyOptPage)

        if self.ui_view is not None:
            self.ui_view.plot_info_view()

    def config_view(self):
        self.stackedWidget.setCurrentWidget(self.configView)
        self.optionStck.setCurrentWidget(self.emptyOptPage)

    def analysis_view(self):
        self.stackedWidget.setCurrentWidget(self.analysisView)
        self.optionStck.setCurrentWidget(self.emptyOptPage)

    def select_single_view(self):
        self.stackedWidget.setCurrentWidget(self.singleView)
        if self.ui_view is not None:
            self.ui_view.plot_single_view()

    def select_grid_view(self):
        self.stackedWidget.setCurrentWidget(self.gridView)
        if self.ui_view is not None:
            self.ui_view.plot_multiple_view()

    def select_list_view(self):
        self.stackedWidget.setCurrentWidget(self.listView)
        if self.ui_view is not None:
            self.ui_view.plot_list_view()

    def extract_spikes(self):
        self.raw_handler.extract_spikes()

    def extract_all_spikes(self):
        self.raw_handler.extract_all_spikes()

    def run_kmeans(self):
        self.spike_handler.run_kmeans()

    def run_all_kmeans(self):
        self.spike_handler.run_all_kmeans()

    def reset_spike_classes(self):
        self.spike_handler.reset_spike_classes()

    def signal_unit_changed(self, changes):
        self.spike_handler.signal_unit_changed(changes)

    def save_spikes(self):
        self.spike_handler.save_spikes()

    def spike_stacked_or_ts(self, val):
        self.plot_view()

    def new_layout(self):
        self.layout_ui.new_layout()

    def delete_layout(self):
        self.layout_ui.delete_layout()

    def set_current_layout(self, selected):
        self.layout_ui.set_current_layout(selected)

    def select_layout(self, selected_name):
        self.layout_ui.select_layout(selected_name)

    def load_layout_names(self, selected_name=None):
        self.layout_ui.load_layout_names(selected_name)

    def plot_layout(self, name):
        self.layout_ui.plot_layout(name)

    def load_analysis(self):
        self.analysis_ui.load_analysis()

    def analysis_selected(self, item, index):
        self.analysis_ui.analysis_selected(item, index)

    def run_analysis(self):
        self.analysis_ui.run_analysis()

    def new_analysis(self):
        self.analysis_ui.new_analysis()

    def delete_analysis(self):
        self.analysis_ui.delete_analysis()

    def edit_analysis(self):
        self.analysis_ui.edit_analysis()

    def export_analysis(self):
        self.analysis_ui.export_analysis()

    def settings_selected(self, item, index):
        self.settings_ui.settings_selected(item, index)

    def settings_save(self):
        self.settings_ui.save()


from PyQt5.QtWidgets import QGridLayout, QVBoxLayout, QHBoxLayout, QWidget, QSizePolicy
from ..plotters.mini_line_2d import MiniLine2D
from PyQt5.QtCore import Qt
from ..utils.logger import Logger

logger = Logger.get_logger()


class GridLine2D:

    def __init__(self, parent):
        layout = QVBoxLayout()
        layout.setSpacing(0)  # Set spacing to zero
        layout.setContentsMargins(0, 0, 0, 0)  # Set margins to zero

        self.channel_plots = {}
        parent.setLayout(layout)
        self.parent = parent

    def setup_grid(self, nrows=8, ncols=15, plot_names=None, show_axis=False):
        if plot_names is None:
            plot_names = [int(i) for i in range(nrows*ncols)]
        while len(plot_names) < nrows*ncols:
            plot_names.append(len(plot_names))
        size_policy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        # Disable updates
        self.parent.setUpdatesEnabled(False)

        # Clear existing widgets from layout
        layout = self.parent.layout()
        while layout.count():
            item = layout.takeAt(0)
            widget = item.widget()
            if widget:
                widget.deleteLater()


        ch_id = 0
        for row in range(nrows):
            row_widget = QWidget()
            row_widget.setSizePolicy(size_policy)

            row_layout = QHBoxLayout(row_widget)
            row_layout.setContentsMargins(0, 0, 0, 0)
            for col in range(ncols):
                if ch_id < len(plot_names):
                    w = QWidget()
                    w.setSizePolicy(size_policy)

                    row_layout.addWidget(w)
                    self.channel_plots[plot_names[ch_id]] = MiniLine2D(w, yaxis=show_axis)

                ch_id += 1
            layout.addWidget(row_widget)
        # Re-enable updates and refresh the layout
        self.parent.setUpdatesEnabled(True)
        self.parent.update()
        return self

    def plot_xy(self, data):
        length = len(data)

        for ix, (ch_id, (time, serie)) in enumerate(data.items()):
            logger.log_process('plotting_grid_lines', ix, length)
            self.channel_plots[ch_id].title(f'{ch_id}').plot_xy(time, serie)
        logger.log_process('plotting_grid_lines', length, length)

    def plot(self, data, *args, **kwargs):
        length = len(data)
        for plt in self.channel_plots.values():
            plt.clear()

        for ix, (ch_id, (values, group)) in enumerate(data.items()):
            logger.log_process('plotting_grid_lines', ix, length)
            self.channel_plots[ch_id].title(f'{ch_id}').plot(values=values, group=group, *args, **kwargs)
        logger.log_process('plotting_grid_lines', length, length)



import unittest
from unittest.mock import patch, MagicMock, call
from neuronautics.ui.abstract_source import AbstractSource


class MockSource(AbstractSource):
    def _plot_single_view(self, path, channel_id):
        pass

    def _plot_grid_view(self, path):
        pass

    def _plot_list_view(self, channel_id):
        pass

    def load_files(self, folder_path):
        return [folder_path]


class TestAbstractSource(unittest.TestCase):
    def test_constructor(self):
        mock_ui = MagicMock()
        data_handler = MagicMock()

        ms = MockSource(mock_ui, data_handler)

        self.assertEqual(ms.ui, mock_ui)
        self.assertEqual(ms.data_handler, data_handler)

    def test_load_files(self):
        mock_ui = MagicMock()
        data_handler = MagicMock()

        ms = MockSource(mock_ui, data_handler)
        f = ms.load_files('random/folder')
        self.assertListEqual(f, ['random/folder'])

    def test_select_view(self):
        mock_ui = MagicMock()
        data_handler = MagicMock()

        ms = MockSource(mock_ui, data_handler)
        ms.filenames = ['file1', 'file2']
        ms.select_view()
        mock_ui.fileSelector.clear.assert_called_once()
        mock_ui.fileSelector.addItem.assert_has_calls([call('file1'), call('file2')])

    def test_select_file(self):
        mock_ui = MagicMock()
        data_handler = MagicMock()

        ms = MockSource(mock_ui, data_handler)
        with (patch.object(ms, 'load_analog_channels_ids') as spy_laci,
              patch.object(ms, 'plot_info_view') as spy_plot_info_view,
              patch.object(ms, 'plot_multiple_view') as spy_plot_multiple_view,
              patch.object(ms, 'plot_list_view') as spy_plot_list_view
              ):
            ms.select_file()

            spy_laci.assert_called_once()
            spy_plot_list_view.assert_called_once()
            spy_plot_info_view.assert_called_once()
            spy_plot_multiple_view.assert_called_once()

    def test_load_analog_channels_ids_no_files(self):
        mock_ui = MagicMock()
        data_handler = MagicMock()

        ms = MockSource(mock_ui, data_handler)
        mock_ui.fileSelector.currentText.return_value = None

        ms.load_analog_channels_ids()
        mock_ui.channelSelector.clear.assert_not_called()

    def test_load_analog_channels_ids(self):
        mock_ui = MagicMock()
        data_handler = MagicMock()

        ms = MockSource(mock_ui, data_handler)
        mock_ui.fileSelector.currentText.return_value = ['f1', 'f2']
        data_handler.return_value.get_analog_channels_ids.return_value = [1,2,3]

        ms.load_analog_channels_ids()
        mock_ui.channelSelector.clear.assert_called_once()
        mock_ui.channelSelector.addItem.assert_has_calls([call('1'), call('2'), call('3')])

    def test_plot_info_view(self):
        mock_ui = MagicMock()
        data_handler = MagicMock()

        ms = MockSource(mock_ui, data_handler)
        data_handler.return_value.get_info.return_value = {'a': 1, 'b': 2}

        with patch('neuronautics.ui.abstract_source.QtWidgets.QTableWidgetItem') as mck_wgt:
            mck_wgt.side_effect = lambda x: x
            ms.plot_info_view()

        mock_ui.infoTable.setItem.assert_has_calls([call(0, 0, 'a'),
                                                    call(0, 1, 1),
                                                    call(1, 0, 'b'),
                                                    call(1, 1, 2), ])

    def test_plot_single_view(self):
        mock_ui = MagicMock()
        data_handler = MagicMock()

        ms = MockSource(mock_ui, data_handler)
        with patch.object(ms, '_plot_single_view') as mck_single_view:
            ms.plot_single_view()

            mck_single_view.assert_called_once()

    def test_plot_grid_view(self):
        mock_ui = MagicMock()
        data_handler = MagicMock()

        ms = MockSource(mock_ui, data_handler)
        with patch.object(ms, '_plot_grid_view') as mck_view:
            ms.plot_multiple_view()

            mck_view.assert_called_once()

    def test_plot_list_view(self):
        mock_ui = MagicMock()
        data_handler = MagicMock()

        ms = MockSource(mock_ui, data_handler)
        with patch.object(ms, '_plot_list_view') as mck_view:
            ms.plot_list_view()

            mck_view.assert_called_once()


if __name__ == '__main__':
    unittest.main()



from PyQt5 import QtWidgets
import sys
from datetime import datetime
from neuronautics import my_resources

from neuronautics.ui.neuronautics_ui import NeuronauticsUi

def main():
    app = QtWidgets.QApplication(sys.argv)

    window = NeuronauticsUi()

    app.exec_()

def extract_spikes(folder):
    green_text = "\033[32m"
    red_text = "\033[91m"
    reset_color = "\033[0m"

    import glob
    import os
    h5_files = glob.glob(os.path.join(folder, '*.h5'))
    file_errors = []
    for ix, file in enumerate(h5_files):
        if not os.path.exists(file.replace('h5', 'spike')):
            print(green_text+f'[{datetime.now()}] {ix+1}/{len(h5_files)} : {file}'+reset_color)
            from neuronautics.recordings.mcs_raw import McsRaw
            try:
                McsRaw(file).extract_all_spikes(5, 1000).to_csv(file.replace('h5', 'spike'), index=False)
            except:
                print(red_text+f'{ix+1}/{len(h5_files)} : {file}'+reset_color)
                file_errors.append(file)
    print('Failing spikes extraction: ')
    for file in file_errors:
        print(f'| {file}')


if __name__ == '__main__':
    main()
    #extract_spikes('/home/dani/OneDrive/MEA')


import os
import numpy as np
import re
from pathlib import Path

import subprocess
import platform
import yaml

from scipy.signal import butter, filtfilt


def butter_lowpass_filter(data, cutoff_freq, sampling_rate, order=4):
    nyquist_freq = 0.5 * sampling_rate
    normal_cutoff = cutoff_freq / nyquist_freq
    b, a = butter(order, normal_cutoff, btype='low', analog=False)
    filtered_data = filtfilt(b, a, data)
    return filtered_data


def moving_stats(array: np.array, window: int = 3) -> (float, float):
    """
    Calculate the moving average and standard deviation of a sequence.

    This function calculates the moving average and moving standard deviation of
    a given sequence using a specified window size.

    Args:
        array (np.array): Input sequence for which the moving average and
            standard deviation need to be calculated.
        window (int, optional): Window size for the moving average and standard
            deviation calculation. Defaults to 3.

    Returns:
        tuple: A tuple containing two arrays representing the calculated
            moving average and moving standard deviation.

    Raises:
        ValueError: If the window size `window` is not a positive integer or if it
            exceeds the length of the input sequence `array`.

    Examples:
        >>> data = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        >>> avg, std = moving_stats(data, window=3)
        >>> print(avg)
        [ 2.  3.  4.  5.  6.  7.  8.  9.]
        >>> print(std)
        [ 0.81649658  0.81649658  0.81649658  0.81649658  0.81649658  0.81649658]
    """

    if (array is None) or (len(array) < window):
        raise ValueError("array must be longer than window size")

    if window < 3:
        raise ValueError("window must be greater than or equal to 3")

    cum_sum = np.cumsum(array, dtype=float)
    cum_sum[window:] = cum_sum[window:] - cum_sum[:-window]

    cum_sum_squares = np.cumsum(array ** 2, dtype=float)
    cum_sum_squares[window:] = cum_sum_squares[window:] - cum_sum_squares[:-window]

    avg = cum_sum[window - 1:] / window
    moving_var = (cum_sum_squares[window - 1:] - cum_sum[window - 1:] ** 2 / window) / window
    std = np.sqrt(moving_var)
    return avg, std


def parse_numeric_array(array_str: str) -> np.array:
    """
    Convert a string representation of an array to a NumPy array.

    This function takes a string representation of an array and extracts the
    numerical values to create a NumPy array.

    Args:
        array_str (str): A string containing numerical values separated by
            non-numeric characters.

    Returns:
        numpy.ndarray: A NumPy array containing the extracted numerical values.

    Raises:
        None

    Examples:
        >>> input_str = "[1.2, 3.4, -5.6, 7.8]"
        >>> result_array = parse_numeric_array(input_str)
        >>> print(result_array)
        array([ 1.2,  3.4, -5.6,  7.8])
    """
    try:
        # Extract the values from the string using regular expressions
        values = re.findall(r'-?\d+\.\d*', array_str)
        # Convert the values to integers and return as an array
        return np.array([float(value) for value in values])
    except (ValueError, TypeError):
        # Handle the case where the string representation is invalid
        return np.array([])


def mkdir(*args):
    """
    Create a directory and its parent directories if they do not exist.

    This function constructs a directory path by joining the provided
    arguments and then creates the directory along with any necessary parent
    directories if they do not already exist.

    Args:
        *args: Variable number of string arguments that represent the path
            segments for the directory to be created.

    Returns:
        str: The path of the created directory.

    Raises:
        None

    Examples:
        >>> new_directory = mkdir('path', 'to', 'my', 'directory')
        >>> print(new_directory)
        'path/to/my/directory'
        >>> existing_directory = mkdir('path', 'to', 'existing', 'directory')
        >>> print(existing_directory)
        'path/to/existing/directory'
    """
    path = '/'.join(args)
    Path(path).mkdir(parents=True, exist_ok=True)
    return path


def app_path(*args):
    """
    Create a path within the application's directory structure.

    This function constructs a path within the application's directory
    structure by appending the provided arguments to the application's base
    directory. The base directory is typically located in the user's home
    directory and is used to organize application-specific files.

    Args:
        *args: Variable number of string arguments that represent the
            subdirectories to be added to the application path.

    Returns:
        str: The path within the application's directory structure.

    Raises:
        None

    Examples:
        >>> app_subpath = app_path('data', 'config')
        >>> print(app_subpath)
        '/home/user/.neuronautics/data/config'
    """
    home_path = os.path.expanduser('~')
    return mkdir(home_path, '.neuronautics', *args)


def file_path(fn: str) -> str:
    """
    Create a full file path within the application's directory structure.

    This function constructs a full file path within the application's directory
    structure by combining the base directory obtained from the `app_path`
    function with the provided file name.

    Args:
        fn (str): The file name or relative file path to be added to the
            application path.

    Returns:
        str: The full file path within the application's directory structure.

    Raises:
        None

    Examples:
        >>> file = 'data/settings.json'
        >>> full_file_path = file_path(file)
        >>> print(full_file_path)
        '/home/user/.neuronautics/data/settings.json'
    """
    fn_path = fn.split('/')
    main_folder = app_path(*fn_path[:-1])
    return '/'.join([main_folder, fn_path[-1]])


def load_yaml(filename, default=None):
    """
    Load YAML data from a file.

    This function reads YAML data from a specified file and returns the loaded
    data as a Python dictionary. If the file does not exist or cannot be read,
    the function returns the specified default value.

    Args:
        filename (str): The path to the YAML file to be loaded.
        default (any, optional): The default value to return if the file does
            not exist or cannot be read. Defaults to None.

    Returns:
        dict or any: A dictionary containing the loaded YAML data if the file
            exists and can be read, or the specified default value otherwise.

    Raises:
        None

    Examples:
        >>> yaml_file = 'config/settings.yaml'
        >>> config_data = load_yaml(yaml_file, default={})
        >>> print(config_data)
        {'key': 'value', 'nested': {'nested_key': 'nested_value'}}
    """
    if Path(filename).exists():
        with open(filename, 'r') as stream:
            data = yaml.full_load(stream)
            if not data:
                return default
            return data
    else:
        return default


def open_external_editor(filename):
    """
    Open a file in an external text editor based on the operating system.

    This function attempts to open a specified file in an external text editor
    based on the user's operating system. It supports Windows, Linux, and macOS.

    Args:
        filename (str): The path to the file to be opened.

    Returns:
        None

    Raises:
        None

    Examples:
        >>> open_external_editor('example.txt')
        # Opens 'example.txt' in the default text editor based on the OS
    """
    system = platform.system()
    editor_commands = {
        "Windows": ["code", "notepad"],
        "Linux": ["code", "xdg-open"],
        "Darwin": ["code", "open -a TextEdit"]
    }

    editors = editor_commands.get(system)
    if editors:
        for editor in editors:
            try:
                subprocess.run([editor, filename], check=True)
                break  # Opened successfully, no need to try other editors
            except subprocess.CalledProcessError:
                pass
        else:
            print("Unable to open the file in any editor.")
    else:
        print("Unsupported operating system.")




class Singleton(type):
    """
    Metaclass implementing the Singleton design pattern.

    This metaclass ensures that a class has only one instance and provides
    a global point of access to that instance.

    Usage:
        class MyClass(metaclass=Singleton):
            # Your class definition

    Attributes:
        _instances (dict): A dictionary storing instances of Singleton classes.

    Methods:
        __call__(cls, *args, **kwargs):
            Override of the call behavior to create and return a single instance
            of the class if it doesn't exist, or return the existing instance.

    Examples:
        class SingletonClass(metaclass=Singleton):
            pass
        instance1 = SingletonClass()
        instance2 = SingletonClass()
        assert instance1 is instance2  # Both references point to the same instance.
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Create and return a single instance of the class if it doesn't exist,
        or return the existing instance.

        Args:
            cls (class): The class being instantiated.
            *args: Positional arguments for the class constructor.
            **kwargs: Keyword arguments for the class constructor.

        Returns:
            object: An instance of the class.
        """
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog, QDialogButtonBox
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QApplication, QWidget, QVBoxLayout, QPushButton, QCheckBox
from neuronautics.plotters.widgets.signal_line import SignalLine


class SignalViewSelector(QDialog):
    NEW = 9999
    CANCEL = -9999

    def __init__(self, signal_ids, signal_visible=None, parent=None):
        super().__init__(parent, Qt.CustomizeWindowHint)

        app = QApplication.instance()
        main_app_stylesheet = app.activeWindow().styleSheet()

        if signal_visible is None:
            signal_visible = [True] * len(signal_ids)

        self.setStyleSheet(main_app_stylesheet)

        widget = QWidget()
        layout = QVBoxLayout(widget)
        self.setLayout(layout)

        # Add buttons to the grid
        signal_ids = sorted(list(signal_ids))
        self.checkbox = {}

        for ix, (signal_id, visible) in enumerate(zip(signal_ids, signal_visible)):
            if signal_id >= 0:
                chk_box = QCheckBox(f"{signal_id}")
                chk_box.setChecked(visible)
                self._style_box(chk_box, QColor(*SignalLine.color(signal_id)))

                layout.addWidget(chk_box)
                self.checkbox[signal_id] = chk_box

        chk_box = QCheckBox(f"Noise")
        chk_box.setChecked(True)
        self._style_box(chk_box, QColor(*SignalLine.color(-1)))
        layout.addWidget(chk_box)
        self.checkbox[-1] = chk_box

        chk_box = QCheckBox(f"Stimulus")
        chk_box.setChecked(True)
        self._style_box(chk_box, QColor(*SignalLine.color(1000)))
        layout.addWidget(chk_box)
        self.checkbox[1000] = chk_box

        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)

        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        layout.addWidget(buttonBox)

        # Set the widget as the message box's layout
        self.layout().addWidget(widget)

    def _style_box(self, chk, background_color):
        style = [f"background-color: {background_color.name()}",
                 "width: 50px",
                 f"color: {background_color.darker(240).name()}",
                 f"border: 2px solid {background_color.darker(120).name()}"]
        chk.setStyleSheet(";".join(style))

    def accept(self):
        self.done(1)

    def reject(self):
        self.done(-1)

    def get_selected(self):
        return {id: chk.isChecked() for id, chk in self.checkbox.items()}



from PyQt5.QtGui import QFont
from PyQt5.QtCore import QMargins
from PyQt5.QtWidgets import QSizePolicy
from ..plotters.line_2d import Line2D


class MiniLine2D(Line2D):
    def __init__(self, parent, yaxis=False, xaxis=False):
        super().__init__(parent)

        self.yaxis.setTickCount(3)  # Adjust the number of ticks
        self.yaxis.setLabelFormat("%.2f")  # Format the labels (adjust as needed)
        # Set font size for the axis labels
        font = QFont()
        font.setPointSize(6)  # Adjust font size
        self.yaxis.setLabelsFont(font)
        self.yaxis.setVisible(yaxis)
        self.xaxis.setVisible(xaxis)

        self.chart.legend().hide()  # Hide the plot legend
        self.chart.layout().setContentsMargins(0, 0, 0, 0)  # Remove margins for the chart
        self.chart.layout().setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding))  # Remove margins for the chart

        self.chart.setBackgroundRoundness(0)  # Set background roundness to zero

        # Set up the chart appearance
        self.chart.setBackgroundVisible(False)  # Hide the chart's background
        self.chart.setPlotAreaBackgroundVisible(False)  # Hide the plot area's background

        # Set the plot area margins to zero to maximize the plot area
        self.chart.setMargins(QMargins(0, 0, 0, 0))
        self.single_chart_view.allow_selection(False)

    def plot_xy(self, xvalues, yvalues):
        super().plot_xy(xvalues, yvalues)
        self.single_chart_view.setContentsMargins(0, 0, 0, 0)  # Remove margins for the QChartView



from abc import ABC

from .abstract_analysis import AbstractAnalysis, AnalysisType
import matplotlib.pyplot as plt


class BarAnalysis(AbstractAnalysis, ABC):

    def type(self):
        return AnalysisType.BAR

    def plot(self, value_units='', stacked=True, vertical=True, title='', *args, **kwargs):
        dataframe = self.run(*args, **kwargs)

        fig, ax = plt.subplots(figsize=(10, 6))

        columns = list(dataframe.columns)

        value_columns = columns[1:]
        label = columns[0]

        kind = 'bar' if vertical else 'barh'
        if stacked:
            dataframe[value_columns].plot(kind=kind, stacked=True, ax=ax)
        else:
            dataframe[value_columns].plot(kind=kind, ax=ax)

        if vertical:
            plt.xticks(range(len(dataframe)), dataframe[label])
            ax.set_ylabel(value_units)
        else:
            plt.yticks(range(len(dataframe)), dataframe[label])
            ax.set_xlabel(value_units)
        plt.legend(loc='center left', bbox_to_anchor=(1.005, 0.5))


        plt.title(title)
        plt.tight_layout()

        return fig



import yaml
from pathlib import Path
from ..utils.helpers import file_path

LAYOUT_FILE = file_path('layout.yml')

class Layout:
    def __init__(self):
        self.layouts = self.load()

    def __getstate__(self):
        return dict()

    def __setstate__(self, state):
        pass

    def save(self):
        with open(LAYOUT_FILE, 'w') as stream:
            yaml.dump(self.layouts, stream)

    def load(self):
        if Path(LAYOUT_FILE).exists():
            with open(LAYOUT_FILE, 'r') as stream:
                data = yaml.full_load(stream)
                if not data:
                    return list()
                return data
        else:
            return list()

    def names(self):
        return sorted([layout.get('name') for layout in self.layouts])

    def new(self, name, cols, rows):
        self.layouts.append({'name': name,
                             'rows': rows, 'cols': cols,
                             'layout': self.default_config(cols, rows),
                             'current': False})
        self.save()

    def _get(self, name):
        for layout in self.layouts:
            if layout.get('name') == name:
                return layout

    def update(self, name, cell_values):
        layout = self._get(name)
        layout['layout'] = cell_values
        self.save()

    def get_layout(self, name):
        layout = self._get(name)
        return layout.get('layout')

    def get_shape(self, name):
        layout = self._get(name)
        return int(layout['cols']), int(layout['rows'])

    @classmethod
    def int_to_abc(cls, n):
        result = ""
        while n > 0:
            n, rem = divmod(n - 1, 26)
            result = chr(65 + rem) + result
        return result

    def default_config(self, cols, rows):
        cell_values = [[f"{self.int_to_abc(c+1)}{r+1}" for c in range(cols)] for r in range(rows)]
        return cell_values

    def column_headers(self, cols):
        return [self.int_to_abc(c+1) for c in range(cols)]

    def row_headers(self, rows):
        return [str(r+1) for r in range(rows)]

    def delete(self, name):
        for ix, layout in enumerate(self.layouts):
            if layout.get('name') == name:
                self.layouts.pop(ix)
                self.save()
                return True
        return False

    def set_current_layout(self, name):
        for layout in self.layouts:
            layout['current'] = layout.get('name') == name
        self.save()

    def is_current_layout(self, name):
        return self._get(name).get('current', False)

    def current(self):
        for layout in self.layouts:
            if layout.get('current', False):
                return layout.get('layout')




from ..utils.singleton import Singleton
from ..config.settings import Settings

settings = Settings()


class SettingsUi(metaclass=Singleton):
    def __init__(self, ui):
        self.ui = ui
        self._mapping_tab = {
            'Raw': self.ui.rawSettings,
            'Spike': self.ui.spikeSettings
        }
        self._config_raw_settings()
        self._config_spike_settings()

    def _config_spike_settings(self):
        self.ui.spikeSingleViewLen.setValue(settings.ui_spike.SINGLE_VIEW_LENGTH_S)
        self.ui.spikeGridStackedViewLen.setValue(settings.ui_spike.GRID_VIEW_STACKED_S)
        self.ui.spikeGridTSViewLen.setValue(settings.ui_spike.GRID_VIEW_TIMESERIES_S)
        if settings.ui_spike.MICRO_VOLTS_RANGE is None:
            self.ui.spikeMinRange.setValue(.0)
            self.ui.spikeMaxRange.setValue(.0)
        else:
            self.ui.spikeMinRange.setValue(settings.ui_spike.MICRO_VOLTS_RANGE[0])
            self.ui.spikeMaxRange.setValue(settings.ui_spike.MICRO_VOLTS_RANGE[1])
        self.ui.spikeSingleViewLen.valueChanged.connect(settings.ui_spike.set_single_view)
        self.ui.spikeGridStackedViewLen.valueChanged.connect(settings.ui_spike.set_grid_stacked_view)
        self.ui.spikeGridTSViewLen.valueChanged.connect(settings.ui_spike.set_grid_timeseries_view)
        self.ui.spikeMinRange.valueChanged.connect(settings.ui_spike.set_volts_min)
        self.ui.spikeMaxRange.valueChanged.connect(settings.ui_spike.set_volts_max)

    def _config_raw_settings(self):
        self.ui.rawSingleViewLen.setValue(settings.ui_raw.SINGLE_VIEW_LENGTH_S)
        self.ui.rawGridViewLen.setValue(settings.ui_raw.GRID_VIEW_LENGTH_S)
        if settings.ui_raw.MICRO_VOLTS_RANGE is None:
            self.ui.rawMinRange.setValue(.0)
            self.ui.rawMaxRange.setValue(.0)
        else:
            self.ui.rawMinRange.setValue(settings.ui_raw.MICRO_VOLTS_RANGE[0])
            self.ui.rawMaxRange.setValue(settings.ui_raw.MICRO_VOLTS_RANGE[1])
        self.ui.rawSingleViewLen.valueChanged.connect(settings.ui_raw.set_single_view)
        self.ui.rawGridViewLen.valueChanged.connect(settings.ui_raw.set_grid_view)
        self.ui.rawMinRange.valueChanged.connect(settings.ui_raw.set_volts_min)
        self.ui.rawMaxRange.valueChanged.connect(settings.ui_raw.set_volts_max)

    def settings_selected(self, item, index):
        selected_item = item.text(0)
        wgt = self.ui.stackedSettings
        wgt.setCurrentWidget(self._mapping_tab.get(selected_item, self.ui.emptySettings))

    def save(self):
        settings.save()


from neuronautics.analysis.type.graph_analysis import GraphAnalysis
from neuronautics.analysis.helpers import to_timeseries
from neuronautics.config.layout import Layout
import numpy as np


class ActivityCorrelation(GraphAnalysis):
    """A class for performing activity correlation analysis.

    This class provides methods for analyzing the correlation of activity
    between different channels based on input spike data.

    Args:
        spikes (DataFrame): A DataFrame containing spike data.
        bin_window_ms (int): The time window (in milliseconds) for binning spikes.
        corr_thr (float): The correlation threshold for considering correlations.

    Attributes:
        spikes (DataFrame): The input spike data.
        bin_window_ms (int): The time window for binning spikes (in milliseconds).
        corr_thr (float): The correlation threshold.

    """

    def get_input_params(self):
        """Get the input parameters for the activity correlation analysis.

        Returns:
            list: A list of dictionaries, each describing an input parameter.
                Each dictionary includes 'name', 'min', 'max', 'default', and 'type' keys.

        """
        return [
            {'name': 'bin_window_ms', 'min': 1, 'max': 1_000, 'default': 100, 'type': 'int'},
            {'name': 'corr_thr', 'min': 0, 'max': 1, 'default': 0.3, 'type': 'float'}
        ]

    def run(self, spikes, bin_window_ms, corr_thr, *args, **kwargs):
        """Run the activity correlation analysis.

        Args:
            spikes (DataFrame): A DataFrame containing spike data.
            bin_window_ms (int): The time window (in milliseconds) for binning spikes.
            corr_thr (float): The correlation threshold for considering correlations.

        Returns:
            ndarray: A boolean matrix indicating correlations above the threshold.

        """
        spikes['class'] = spikes['class'].astype(int)
        spikes = spikes[spikes['class'] >= 0]  # remove noise
        spikes = spikes[['channel_id', 'ts_ms']].copy().reset_index(drop=True)

        bin_df = to_timeseries(spikes, group_lvl='channel_id', bin_ms=bin_window_ms)

        layout = Layout().current()
        max_lay = np.max([np.max([int(l) for lay in layout for l in lay if l != ''])])
        w_ij = np.zeros((max_lay, max_lay))
        for i, (ch_id_i, bin_i, events_i) in bin_df.iterrows():
            max_i = events_i.sum()
            for j, (ch_id_j, bin_j, events_j) in bin_df.iterrows():
                max_j = events_j.sum()
                shift_0 = (events_i[:-1] * events_j[1:]).sum()
                shift_1 = (events_i[1:] * events_j[:-1]).sum()
                shift_2 = (events_i * events_j).sum()
                w_ij[ch_id_i, ch_id_j] = (shift_0 + shift_1 + shift_2) / (3 * max(max_i, max_j))
                w_ij[ch_id_i, ch_id_j] = 0 if ch_id_i == ch_id_j else w_ij[ch_id_i, ch_id_j]

        return w_ij > corr_thr
    
    def plot(self, *args, **kwargs):
        """Plot the activity correlation analysis.

        Args:
            *args: Variable-length positional arguments.
            **kwargs: Variable-length keyword arguments.

        Returns:
            object: matplotlib figure

        """
        return super().plot('Activity Correlation', *args, **kwargs)



import glob

import numpy as np

from ..recordings.nn_spike import NNSpike

from ..mlkit.spike_sorter import SpikeSorter

from ..utils.logger import Logger
from ..ui.abstract_source import AbstractSource
from ..config.settings import Settings
settings = Settings()

logger = Logger.get_logger()


class Spike(AbstractSource):
    """
    A class representing spike data visualization and processing.

    This class handles the visualization and processing of spike data, including
    loading files, plotting different views, running k-means clustering, saving data,
    and managing signal unit changes.

    Methods:
        __init__(self, ui): Constructor to initialize the Spike object.
        load_files(self, folder_path): Load spike files from the given folder path.
        select_view(self): Select and display the spike view.
        _plot_single_view(self, path, channel_id): Plot single view of spike data.
        _plot_grid_view(self, path): Plot grid view of spike data.
        _plot_list_view(self, channel_id): Plot list view of spike data.
        all_data(self): Get all spike data.
        run_kmeans(self): Perform k-means clustering on spike data.
        run_all_kmeans(self): Perform k-means clustering on all spike data channels.
        reset_spike_classes(self): Reset spike classes to default.
        signal_unit_changed(self, changes): Update spike unit classes based on changes.
        save_spikes(self): Save spike data.
    """
    def __init__(self, ui):
        super().__init__(ui, NNSpike)

    def load_files(self, folder_path):
        """
        Load spike files from the given folder path.

        Args:
            folder_path (str): Path to the folder containing spike files.
        """
        self.ui.spikeBtn.setEnabled(False)

        self.filenames = sorted(glob.glob(f"{folder_path}/*.spike"))

        if len(self.filenames) > 0:
            self.ui.spikeBtn.setEnabled(True)

    def select_view(self):
        """
        Selects and displays the spike view.

        This method is called to select the spike view and update the user interface
        accordingly. It sets the current widget of the option stack to the spike
        options page.

        Usage:
            Call this method when you want to switch to the spike view.
        """
        super().select_view()
        self.ui.optionStck.setCurrentWidget(self.ui.spikeOptPage)

    def _plot_single_view(self, path, channel_id):
        as_timeseries = 1-int(self.ui.chkSpikeStacked.isChecked())
        nn = self.data_handler(path)
        start_ms = int(self.ui.recording_start_s) * 1_000
        if as_timeseries == 1:
            end_ms = int(self.ui.recording_start_s + settings.ui_spike.SINGLE_VIEW_LENGTH_S) * 1_000
            xlim = (start_ms-1, end_ms)
        else:
            end_ms = np.Inf
            xlim = None
        channel_data = nn.get_channel_data(channel_id, start_ms, end_ms)

        values, group = [], []
        for ix, (cl, t, spk) in channel_data.iterrows():
            values.append([(t * as_timeseries + nn.to_ms(x, 'tick') - 1, y) for x, y in enumerate(spk)])
            group.append(cl)

        (
            self.ui.single_chart_view
            .x_label('Time (ms)')
            .y_label('Microvolts')
            .title(f'Channel {channel_id}')
            .plot(values=values, group=group, allow_selection=True,
                  xlim=xlim, ylim=settings.ui_spike.MICRO_VOLTS_RANGE,
                  progress_bar=True)
        )

    def _plot_grid_view(self, path):
        nn = self.data_handler(path)
        as_timeseries = 1-int(self.ui.chkSpikeStacked.isChecked())

        start_ms = int(self.ui.recording_start_s) * 1_000

        if as_timeseries == 1:
            end_ms = int((self.ui.recording_start_s + settings.ui_spike.GRID_VIEW_TIMESERIES_S) * 1_000)
            xlim = (start_ms-1, end_ms)
        else:
            end_ms = int((self.ui.recording_start_s + settings.ui_spike.GRID_VIEW_STACKED_S) * 1_000)
            xlim = None

        channel_data = nn.get_all_data( start_ms, end_ms)

        channel_values = dict()
        for ix, (ch_id, cl, t, spk) in channel_data.iterrows():
            values, group = channel_values.get(ch_id, ([], []))
            values.append([(t * as_timeseries + nn.to_ms(x, 'tick') - 1, y) for x, y in enumerate(spk)])
            group.append(cl)
            channel_values[ch_id] = (values, group)

        (
            self.ui.multiple_chart_view
            .plot(channel_values, xlim=xlim, ylim=settings.ui_spike.MICRO_VOLTS_RANGE)
        )

    def _plot_list_view(self, channel_id):
        as_timeseries = 1-int(self.ui.chkSpikeStacked.isChecked())
        start_ms = int(self.ui.recording_start_s) * 1_000

        if as_timeseries == 1:
            end_ms = int((self.ui.recording_start_s + settings.ui_spike.GRID_VIEW_TIMESERIES_S) * 1_000)
            xlim = (start_ms-1, end_ms)
        else:
            end_ms = int((self.ui.recording_start_s + settings.ui_spike.GRID_VIEW_STACKED_S) * 1_000)
            xlim = None

        channel_values = dict()
        for ix, fn in enumerate(self.filenames):

            logger.log_process('_plot_list_view', ix, len(self.filenames))
            name = fn.split('/')[-1].replace('.spike', '')
            nn = self.data_handler(fn)

            channel_data = nn.get_channel_data(channel_id, start_ms, end_ms)

            values, group = [], []
            for ix, (cl, t, spk) in channel_data.iterrows():
                values.append([(t * as_timeseries + nn.to_ms(x, 'tick') - 1, y) for x, y in enumerate(spk)])
                group.append(cl)
            channel_values[name] = (values, group)

        logger.log_process('_plot_list_view', len(self.filenames), len(self.filenames))

        self.ui.list_chart_view.plot(channel_values, xlim=xlim, ylim=settings.ui_spike.MICRO_VOLTS_RANGE)

    def all_data(self):
        """
        Retrieves all spike data from the currently selected file.

        This method retrieves all spike data from the currently selected file
        using the NNSpike class and returns it as a DataFrame.

        Returns:
            pandas.DataFrame: A DataFrame containing all spike data.

        Usage:
            spike_instance = Spike(ui)
            data_frame = spike_instance.all_data()
        """
        path = self.ui.fileSelector.currentText()
        nn = self.data_handler(path)
        df = nn.get_all_data(0, np.Inf)
        return df

    def all_channel_data(self):
        """
        Retrieves all spike data from the currently selected file and channel.

        This method retrieves all spike data from the currently selected file and channel
        using the NNSpike class and returns it as a DataFrame.

        Returns:
            pandas.DataFrame: A DataFrame containing all spike data.

        Usage:
            spike_instance = Spike(ui)
            data_frame = spike_instance.all_channel_data()
        """
        path = self.ui.fileSelector.currentText()
        channel_id = int(self.ui.channelSelector.currentText())
        nn = self.data_handler(path)
        df = nn.get_channel_data(channel_id, 0, np.Inf)
        return df

    def run_kmeans(self):
        """
        Runs k-means clustering on spike data for the current channel.

        This method performs k-means clustering on the spike data for the current
        channel using the specified number of principal components and clusters.
        It then sets the computed labels for the channel using the NNSpike class.

        Usage:
            spike_instance = Spike(ui)
            spike_instance.run_kmeans()

        Notes:
            This method requires a valid file selection and channel choice in the UI.
        """
        stim_up, stim_down, stim_shift = self._stimulus_params()
        path, channel_id, num_pca, num_clusters = self._get_kmeans_params()

        df = self.all_channel_data()

        labels = self._pca_kmeans(df, num_pca, num_clusters)
        labels = self._relabel(df, labels, stim_up, stim_down, stim_shift)

        labels = [l2 if l2 == -1 else l1 for (l1, l2) in zip(labels, df['class'])]

        nn = self.data_handler(path)
        nn.set_labels(channel_id, labels)

        self.plot_single_view()

    def _stimulus_params(self):
        stim_up = self.ui.stimUp.value()
        stim_down = self.ui.stimDown.value()
        stim_shift = self.ui.stimShift.value()
        return stim_up, stim_down, stim_shift

    def _get_kmeans_params(self):
        path = self.ui.fileSelector.currentText()
        channel_id = int(self.ui.channelSelector.currentText())
        num_pca = self.ui.pcaValue.value()
        num_clusters = self.ui.kValue.value()
        return path, channel_id, num_pca, num_clusters

    def _pca_kmeans(self, df, num_pca, num_clusters):
        # Extract spike data for the current channel
        spike_data = np.array(df['spike'].tolist())
        labels = (
            SpikeSorter(spike_data)
            .detect_noise()
            .pca(num_pca)
            .kmeans(num_clusters)
            .run()
        )

        return labels

    def _relabel(self, df, labels, stim_up, stim_down, stim_shift):
        snapshot = list(df.ts_ms)
        i = stim_shift
        lab = []
        for s in snapshot:
            l = 0
            while s >= i + stim_up:
                i += stim_up + stim_down
            if s >= i and s < i + stim_up:
                l = -1
            lab.append(l)
        if len(lab) == len(labels):
            return [l2 if l2 == -1 else l1 for (l1,l2) in zip(labels, lab)]
        else:
            return labels

    def run_all_kmeans(self):
        """
        Runs k-means clustering on spike data for all channels.

        This method performs k-means clustering on the spike data for all channels
        using the specified number of principal components and clusters. It iterates
        through each channel, computes labels, and sets them using the NNSpike class.

        Usage:
            spike_instance = Spike(ui)
            spike_instance.run_all_kmeans()

        Notes:
            This method requires a valid file selection.
        """
        stim_up, stim_down, stim_shift = self._stimulus_params()
        path, _, num_pca, num_clusters = self._get_kmeans_params()

        nn = self.data_handler(path)
        df = nn.get_all_data(0, np.Inf)

        logger = Logger.get_logger()

        grouped = df.groupby('channel_id')
        n_groups = len(grouped)
        logger.log_process('run_all_kmeans', 0, n_groups)
        for ix, (channel_id, group) in enumerate(grouped):
            logger.log_process('run_all_kmeans', ix, n_groups)
            labels = self._pca_kmeans(group, num_pca, num_clusters)
            labels = self._relabel(group, labels, stim_up, stim_down, stim_shift)

            labels = [l2 if l2 == -1 else l1 for (l1, l2) in zip(labels, group['class'])]
            nn.set_labels(channel_id, labels)

        logger.log_process('run_all_kmeans', n_groups, n_groups)

        self.plot_single_view()
        self.plot_multiple_view()

    def reset_spike_classes(self):
        """
        Resets spike classes to the default value for the current channel.

        This method resets the spike classes for the current channel to the default
        value (0) using the NNSpike class. It effectively clears any clustering
        labels that were previously applied.

        Usage:
            spike_instance = Spike(ui)
            spike_instance.reset_spike_classes()

        Notes:
            This method requires a valid file selection and channel choice in the UI.
        """
        path, channel_id, _, _ = self._get_kmeans_params()

        nn = self.data_handler(path)
        df = nn.get_channel_data(channel_id, 0, np.Inf)

        nn.set_labels(channel_id, [0] * len(df))

        self.plot_single_view()

    def signal_unit_changed(self, changes):
        """
        Update spike unit classes based on changes.

        This method updates the spike unit classes based on the provided changes
        using the NNSpike class. It merges the changes with the existing spike data
        and updates the labels accordingly.

        Args:
            changes (list of tuples): A list of tuples containing timestamp (in ms)
                                      and new unit class values.

        Usage:
            spike_instance = Spike(ui)
            changes = [(timestamp_1, new_class_1), (timestamp_2, new_class_2)]
            spike_instance.signal_unit_changed(changes)

        Notes:
            This method requires a valid file selection and channel choice in the UI.
        """
        path, channel_id, _, _ = self._get_kmeans_params()

        nn = self.data_handler(path)
        df = nn.get_channel_data(channel_id, 0, np.Inf)

        old_class = list(df['class'])
        for ix, new_class in changes:
            old_class[ix] = new_class

        nn.set_labels(channel_id, old_class)

    def save_spikes(self):
        """
        Save spike data to the original file.

        This method saves the modified spike data to the original file using the
        NNSpike class. It ensures that any changes made to the spike data, such as
        clustering labels, are persisted in the file.

        Usage:
            spike_instance = Spike(ui)
            spike_instance.save_spikes()

        Notes:
            This method requires a valid file selection.
        """
        path = self.ui.fileSelector.currentText()
        logger.log_process('save_spikes', 1, 2)

        self.data_handler(path).save()
        logger.log_process('save_spikes', 2, 2)




import unittest
from unittest.mock import mock_open, patch
from neuronautics.analysis.analysis_creator import AnalysisCreator, AnalysisCreatorUi
from PyQt5 import QtWidgets



class TestAnalysisCreator(unittest.TestCase):
    def test_create(self):
        # Define test input data
        name = "Test Analysis"
        type = "graph"

        # Mock the behavior of open() to return a predefined template
        with (patch("builtins.open", mock_open(read_data="template content")) as mock_file_open,
              patch("neuronautics.analysis.analysis_creator.AnalysisCreator.save") as spy_save,
              patch("neuronautics.analysis.analysis_creator.file_path") as spy_file_path
              ):
            spy_file_path.side_effect = lambda x: f'/home/mock/{x}'
            # Call the create method
            result = AnalysisCreator.create(name, type)

        # Assert that the file was written with the correct content
        mock_file_open.assert_called_with("/home/mock/analysis/custom/scripts/graph/test_analysis.py", "w")
        mock_file_open.return_value.write.assert_called_once_with("template content")

        spy_save.assert_called_once()
        # Assert that the result is the expected path
        self.assertEqual(result, "/home/mock/analysis/custom/scripts/graph/test_analysis.py")

    def test_save(self):
        # Define test input data
        name = "Test Analysis"
        path = "../path/to/your/generated/analysis/graph_test_analysis.py"
        module = "graph_test_analysis"
        class_name = "GraphTestAnalysis"

        # Mock the Loader.save method to ensure it's called with the correct arguments
        with patch("neuronautics.analysis.analysis_creator.Loader.save") as mock_loader_save:
            AnalysisCreator.save(name, path, module, class_name)

        # Assert that Loader.save was called with the correct arguments
        mock_loader_save.assert_called_once_with({
            'name': name,
            'module': module,
            'path': path,
            'class': class_name
        })


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pandas as pd


def to_timeseries(data: pd.DataFrame, group_lvl:str, bin_ms: int = 100) -> pd.DataFrame:
    """
    Binarize a time series of events based on a specified bin size.

    This function takes a DataFrame containing timestamped events and
    converts it into a binarized time series. Each event is assigned to a
    bin, and within each bin, a binary array is created to represent the
    presence or absence of events. The resulting time series is grouped by
    a specified grouping level.

    Args:
        data (DataFrame): A DataFrame containing timestamped events.
        group_lvl (str): The column name used for grouping the time series.
        bin_ms (int, optional): The time window (in milliseconds) for binning
            the events. Default is 100 ms.

    Returns:
        DataFrame: A DataFrame containing the binarized time series with
            columns 'group_lvl' and 'events', where 'events' contains binary
            arrays representing event occurrences within each bin.

    Example:
        Given input data like this:

        ```
          ts_ms  channel_id
        0    100           1
        1    200           2
        2    300           1
        3    400           3
        4    500           2
        ```

        Calling `to_timeseries(data, 'channel_id', bin_ms=100)` would produce
        the following output:

        ```
          channel_id                        events
        0           1  [1, 1, 0, 0, 0]
        1           2  [0, 0, 1, 0, 1]
        2           3  [1, 0, 0, 0, 0]
        ```

    """
    data['bin'] = (data['ts_ms'] // bin_ms).astype(int)
    end_time = data['bin'].max()

    bin_df = data.groupby(group_lvl)['bin'].agg(list).reset_index()
    def _to_timeseries(events):
        bin = np.zeros(end_time + 1)
        bin[events] = 1
        return bin

    bin_df['events'] = bin_df.bin.map(_to_timeseries)
    return bin_df






from jinja2 import Template

from ..utils.helpers import file_path, open_external_editor
from ..analysis.loader import Loader
from ..recordings.config import (UI_ANALYSIS_CREATION, JINJA_BASE_ANALYSIS, JINJA_GRAPH_ANALYSIS, JINJA_BAR_ANALYSIS,
                                 JINJA_IMAGE_ANALYSIS, JINJA_LINE_ANALYSIS)


# pyqt
from PyQt5 import QtWidgets, uic
from PyQt5.QtCore import Qt


class AnalysisCreatorUi(QtWidgets.QDialog):
    """
    Class for creating a user interface for analysis creation.

    Attributes:
        None

    Methods:
        __init__: Initializes the AnalysisCreatorUi class.
        create_analysis: Creates an analysis based on user input.
    """
    def __init__(self):
        super(AnalysisCreatorUi, self).__init__()
        self.setWindowFlags(Qt.CustomizeWindowHint | Qt.WindowCloseButtonHint)

        uic.loadUi(UI_ANALYSIS_CREATION, self)

        self.show()

    def create_analysis(self):
        name = self.analysisName.text()
        type = self.analysisType.currentText()

        path = AnalysisCreator.create(name, type)

        self.done(1)
        open_external_editor(path)


class AnalysisCreator:
    """
    Class for creating analysis scripts.

    Attributes:
        None

    Methods:
        create: Creates an analysis script based on the provided name and type.
        save: Saves information about the created analysis.
    """

    @classmethod
    def create(cls, name, analysis_type):
        """
        Creates an analysis script based on the provided name and type.

        Args:
            name (str): The name of the analysis.
            analysis_type (str): The type of analysis to create.

        Returns:
            str: The path to the created analysis script.
        """
        jinja_template = dict(
            graph=JINJA_GRAPH_ANALYSIS,
            bar=JINJA_BAR_ANALYSIS,
            line=JINJA_LINE_ANALYSIS,
            image=JINJA_IMAGE_ANALYSIS
        ).get(analysis_type, JINJA_BASE_ANALYSIS)

        with open(jinja_template) as file_:
            template = Template(file_.read())
            name_split = name.lower().split()
            filename = '_'.join(name_split)
            class_name = ''.join([n.capitalize() for n in name_split])

            # Render the template with the provided values
            filled_template = template.render(
                class_name=class_name,
                title=name
            )

            # Write the filled template to a .py file
            module = ['analysis', 'custom', 'scripts', analysis_type, filename]
            relative_path = '/'.join(module)+'.py'
            path = file_path(relative_path)
            with open(path, 'w') as file:
                file.write(filled_template)

            cls.save(name, relative_path, filename, class_name)

            print(f"Python script '{filename}' generated successfully.")
            return path

    @classmethod
    def save(cls, name, path, module, class_name):
        """
        Saves information about the created analysis.

        Args:
            name (str): The name of the analysis.
            path (str): The path to the created analysis script.
            module (str): The module of the analysis.
            class_name (str): The name of the analysis class.

        Returns:
            None
        """
        config = {
            'name': name,
            'module': module,
            'path': path,
            'class': class_name
        }
        Loader.save(config)



from abc import ABC

from neuronautics.analysis.type.abstract_analysis import AbstractAnalysis, AnalysisType
from neuronautics.analysis.type.palette import Palette
import matplotlib.pyplot as plt


class ImageAnalysis(AbstractAnalysis, ABC):

    def type(self):
        return AnalysisType.IMAGE

    def plot(self, xlabel='', ylabel='', title='', cmap=None, *args, **kwargs):
        matrix = self.run(*args, **kwargs)

        fig, ax = plt.subplots(figsize=(10, 6))

        if cmap:
            ax = ax.imshow(matrix, cmap=cmap)
            plt.colorbar(ax)
        else:
            ax.imshow(matrix)

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(title)
        plt.grid(False)

        return fig



import unittest
from unittest.mock import patch, MagicMock
from neuronautics.analysis.type.line_analysis import LineAnalysis
from neuronautics.analysis.type.abstract_analysis import AnalysisType
from matplotlib.figure import Figure
import pandas as pd
import numpy as np


class OneLineAnalysis(LineAnalysis):

    def run(self, *args, **kwargs):
        return pd.DataFrame({'A': [1,2], 'B': [3,4]})

    def get_input_params(self):
        pass


class TestBarAnalysis(unittest.TestCase):
    def test_type(self):
        analysis = OneLineAnalysis()
        self.assertEqual(analysis.type(), AnalysisType.LINE)

    @patch("matplotlib.pyplot.subplots")
    def test_plot(self, mock_subplots):
        mock_fig = MagicMock()
        mock_ax = MagicMock()
        mock_subplots.return_value = (mock_fig, mock_ax)
        # Act
        analysis = OneLineAnalysis()
        result = analysis.plot(title='my plot')

        # Assert
        self.assertEqual(result, mock_fig)
        mock_ax.plot.assert_called_once()


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import patch, MagicMock, call
from neuronautics.ui.raw import Raw
from neuronautics.ui.abstract_source import AbstractSource
from neuronautics.recordings.mcs_raw import McsRaw
from neuronautics.config.settings import Settings
settings = Settings()

import pandas as pd
import numpy as np


class TestRaw(unittest.TestCase):
    def test_constructor(self):
        mock_ui = MagicMock()
        raw = Raw(ui=mock_ui)
        self.assertIsInstance(raw, AbstractSource)
        self.assertIs(raw.ui, mock_ui)
        self.assertIs(raw.data_handler, McsRaw)

    def test_load_files(self):
        mock_ui = MagicMock()
        files = ['file1.h5', 'file2.h5']
        with patch('glob.glob') as spy_glob:
            spy_glob.return_value = files
            raw = Raw(mock_ui)
            raw.load_files('/random/folder')

            call_args_list = mock_ui.rawBtn.setEnabled.call_args_list
            self.assertEqual(call_args_list[0].args, (False,))
            self.assertEqual(call_args_list[1].args, (True,))
            self.assertListEqual(raw.filenames, files)

    def test_load_files_empty_folder(self):
        mock_ui = MagicMock()
        files = []
        with patch('glob.glob') as spy_glob:
            spy_glob.return_value = files
            raw = Raw(mock_ui)
            raw.load_files('/random/folder')

            mock_ui.rawBtn.setEnabled.assert_called_once_with(False)
            self.assertListEqual(raw.filenames, [])

    def test_select_view(self):
        with patch('neuronautics.ui.spike.AbstractSource.select_view') as spy_select_view:
            mock_ui = MagicMock()
            raw = Raw(mock_ui)

            raw.select_view()

            spy_select_view.assert_called_once()
            mock_ui.optionStck.setCurrentWidget.assert_called_once_with(mock_ui.rawOptPage)

    def test_plot_single_view(self):
        mock_ui = MagicMock()
        mock_ui.recording_start_s = 0
        mock_ui.single_chart_view.x_label.return_value = mock_ui.single_chart_view
        mock_ui.single_chart_view.y_label.return_value = mock_ui.single_chart_view
        mock_ui.single_chart_view.title.return_value = mock_ui.single_chart_view
        mock_ui.single_chart_view.plot_xy.return_value = mock_ui.single_chart_view

        spy_data_handler = MagicMock()
        spy_data_handler.return_value = spy_data_handler
        spy_data_handler.to_ticks.side_effect = lambda x, _: x
        spy_data_handler.get_channel_data.return_value = ([0, 1, 2], [0.1, 0.2, 0.3])

        spike = Raw(mock_ui)
        spike.data_handler = spy_data_handler

        spike._plot_single_view(path='/random/file.h5', channel_id=0)

        mock_ui.single_chart_view.x_label.assert_called_once_with('Time (seconds)')
        mock_ui.single_chart_view.y_label.assert_called_once_with('Microvolts')
        mock_ui.single_chart_view.title.assert_called_once_with('Channel 0')
        mock_ui.single_chart_view.plot_xy.assert_called_once_with(
            [0, 1, 2], [0.1, 0.2, 0.3],
            ylim=settings.ui_raw.MICRO_VOLTS_RANGE
        )

    def test_plot_grid_view(self):
        mock_ui = MagicMock()
        mock_ui.recording_start_s = 0
        mock_ui.multiple_chart_view.plot_xy.return_value = mock_ui.multiple_chart_view

        spy_data_handler = MagicMock()
        spy_data_handler.return_value = spy_data_handler
        spy_data_handler.to_ticks.side_effect = lambda x, _: x
        spy_data_handler.get_all_data.return_value = {0: ([0, 1, 2], [0.1, 0.2, 0.3])}

        spike = Raw(mock_ui)
        spike.data_handler = spy_data_handler

        spike._plot_grid_view(path='/random/file.h5')

        mock_ui.multiple_chart_view.plot_xy.assert_called_once_with(
            {0: ([0, 1, 2], [0.1, 0.2, 0.3])},
            ylim=settings.ui_spike.MICRO_VOLTS_RANGE
        )

    def test_plot_list_view(self):
        mock_ui = MagicMock()
        mock_ui.recording_start_s = 0
        mock_ui.list_chart_view.plot_xy.return_value = mock_ui.list_chart_view

        spy_data_handler = MagicMock()
        spy_data_handler.return_value = spy_data_handler
        spy_data_handler.to_ticks.side_effect = lambda x, _: x
        spy_data_handler.get_channel_data.return_value = ([0, 1, 2], [0.1, 0.2, 0.3])

        spike = Raw(mock_ui)
        spike.filenames = ['file.h5']
        spike.data_handler = spy_data_handler

        spike._plot_list_view(channel_id=1)

        mock_ui.list_chart_view.plot_xy.assert_called_once_with(
            {'file': ([0, 1, 2], [0.1, 0.2, 0.3])}
        )

    def test_get_filename(self):
        fn = Raw.get_filename('/path/to/file.h5')
        self.assertEqual(fn, 'file')

    def test_extract_spikes(self):
        mock_ui = MagicMock()
        mock_ui.fileSelector.currentText.return_value = 'file.h5'
        mock_ui.sigmaValue.value.return_value = 4
        mock_ui.windowValue.value.return_value = 1000

        spy_data_handler = MagicMock()
        spy_data_handler.return_value = spy_data_handler
        spy_data_handler.extract_all_spikes.return_value = MagicMock()

        spike = Raw(mock_ui)
        spike.data_handler = spy_data_handler

        spike.extract_spikes()

        spy_data_handler.extract_all_spikes.assert_called_once_with(4, 1000)
        spy_data_handler.extract_all_spikes.return_value.to_csv.assert_called_once_with('file.spike', index=False)

    def test_extract_spikes_empty_selector(self):
        mock_ui = MagicMock()
        mock_ui.fileSelector.currentText.return_value = None
        mock_ui.sigmaValue.value.return_value = 4
        mock_ui.windowValue.value.return_value = 1000

        spy_data_handler = MagicMock()
        spy_data_handler.return_value = spy_data_handler
        spy_data_handler.extract_all_spikes.return_value = MagicMock()

        spike = Raw(mock_ui)
        spike.data_handler = spy_data_handler

        spike.extract_spikes()

        spy_data_handler.extract_all_spikes.assert_not_called()

    def test_extract_all_spikes(self):
        mock_ui = MagicMock()
        mock_ui.fileSelector.count.return_value = 2

        with patch('PyQt5.QtWidgets.QApplication.processEvents') as spy_events:
            spike = Raw(mock_ui)
            spike.extract_spikes = MagicMock()

            spike.extract_all_spikes()

            mock_ui.fileSelector.setCurrentIndex.assert_has_calls([call(0), call(1)])
            spike.extract_spikes.assert_called()


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import patch
from neuronautics.analysis.correlation.spike_time_tiling_coefficient import SpikeTimeTilingCoefficient
import pandas as pd
import numpy as np


class TestSpikeTimeTilingCoefficient(unittest.TestCase):
    def test_get_input_params(self):
        ac = SpikeTimeTilingCoefficient()
        params = ac.get_input_params()

        for param in params:
            self.assertTrue(param['default'] <= param['max'])
            self.assertTrue(param['min'] <= param['default'])
            self.assertTrue('name' in param)
            self.assertTrue('type' in param)

    def test_run(self):
        spikes = pd.DataFrame({'class': [0, 0, 0, 0, 0, 0, 0, 0],
                               'channel_id': [0, 0, 0, 0, 1, 1, 1, 1],
                               'ts_ms': [1, 2, 3, 4, 1, 2, 3, 4]})
        expected_correlation = np.array([[False, True], [True, False]])

        with patch('neuronautics.analysis.correlation.spike_time_tiling_coefficient.Layout') as mck_layout:
            mck_layout.return_value.current.return_value = [[1, 2]]
            ac = SpikeTimeTilingCoefficient()
            output = ac.run(spikes, 0.5)

            np.testing.assert_array_equal(expected_correlation, output)

    def test_plot(self):
        """parent plot is called with the plot title set"""
        with patch('neuronautics.analysis.correlation.spike_time_tiling_coefficient.super') as spy_super:
            ac = SpikeTimeTilingCoefficient()
            ac.plot()
            spy_super().plot.assert_called_once_with('Spike Time Tiling Coefficient')



if __name__ == '__main__':
    unittest.main()



import pandas as pd
import numpy as np
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans


class SpikeSorter:

    def __init__(self, data):
        self.data = data
        self.output = None
        self.noise = None

    def detect_noise(self):
        self.noise = [all(spike < 0) or all(spike > 0) for spike in self.data]
        return self

    def pca(self, num_components):
        pca = PCA(n_components=num_components)
        self.data = pca.fit_transform(self.data)
        return self

    def kmeans(self, num_clusters):
        # Apply K-means clustering
        kmeans = KMeans(n_clusters=num_clusters, n_init='auto')
        labels = kmeans.fit_predict(self.data)
        self.output = [int(lab) + 1 for lab in labels]
        if self.noise is not None:
            self.output = [-1 if noise else lab for noise, lab in zip (self.noise, self.output)]
        return self

    def run(self):
        return self.output





def clear_layout(layout):
    """
    Clears all widgets from the given QLayout.

    Args:
        layout (QLayout): The QLayout to be cleared.

    Returns:
        None
    """
    while layout.count():
        item = layout.takeAt(0)
        widget = item.widget()
        if widget:
            widget.deleteLater()



import unittest
from neuronautics.analysis.helpers import to_timeseries
import pandas as pd
import numpy as np


class TestHelpers(unittest.TestCase):

    def test_to_timeseries(self):
        # Create sample data
        data = pd.DataFrame({
            'ts_ms': [100, 200, 300, 400, 500],
            'channel_id': [1, 2, 1, 3, 2]
        })

        group_lvl = 'channel_id'
        bin_ms = 100

        result = to_timeseries(data, group_lvl, bin_ms)

        expected_result = pd.DataFrame(
            {'channel_id': [1, 2, 3],
             'bin': [[1, 3], [2, 5], [4]],
             'events': [[0., 1., 0., 1., 0., 0.],
                        [0., 0., 1., 0., 0., 1.],
                        [0., 0., 0., 0., 1., 0.]]}
        )

        pd.testing.assert_frame_equal(expected_result, result)


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import patch, MagicMock

from neuronautics.analysis.loader import Loader


class TestLoader(unittest.TestCase):
    def setUp(self):
        self.loader = Loader()

    @patch('neuronautics.analysis.loader.load_yaml')
    def test_load(self, mock_load_yaml):
        # Setup
        Loader._instance_analysis = MagicMock(return_value=('test', 'class'))
        mock_load_yaml.return_value = [{'name': 'test'}]
        # Call
        self.loader.load()
        # Assert
        self.assertEqual(self.loader.analysis_config, {'test': {'name': 'test'}})
    @patch('neuronautics.analysis.loader.load_yaml')
    def test_load_2(self, mock_load_yaml):
        # Setup
        Loader._instance_analysis = MagicMock(return_value=('test', 'class'))
        mock_load_yaml.side_effect = [[{'name': 'test'}], [{'name': 'othertest'}]]
        # Call
        self.loader.load()
        # Assert
        self.assertEqual(self.loader.analysis_config, {'othertest': {'name': 'othertest'},
                                                       'test': {'name': 'test'}})

    @patch('neuronautics.analysis.loader.file_path')
    def test_get_analysis_path(self, mock_file_path):
        # Setup
        mock_file_path.return_value = 'mymockedpath'
        self.loader.analysis_config = {'test': {'path': 'test_path'}}
        # Call
        result = self.loader.get_analysis_path('test')
        # Assert
        self.assertEqual(result, 'mymockedpath')

    @patch('neuronautics.analysis.loader.file_path')
    def test_get_analysis_path_none(self, mock_file_path):
        # Setup
        mock_file_path.return_value = 'mymockedpath'
        self.loader.analysis_config = {'test': {}}
        # Call
        result = self.loader.get_analysis_path('test')
        # Assert
        self.assertEqual(result, None)

    @patch('neuronautics.analysis.loader.Loader._get_module')
    def test__instance_analysis(self, mock_get_module):
        # Setup
        mock_module = MagicMock()
        mock_get_module.return_value = mock_module
        mock_module.TestClass = MagicMock()
        # Call
        result = self.loader._instance_analysis({'name': 'test', 'module': 'test_module', 'class': 'TestClass', 'path': 'test_path'})
        # Assert
        self.assertEqual(result, ('test', mock_module.TestClass()))

    def test__instance_analysis_exception(self):
        # Call
        with self.assertRaises(Exception):
            result = self.loader._instance_analysis({'name': 'test', 'module': 'test_module', 'class': 'TestClass', 'path': 'test_path'})

    @patch('neuronautics.analysis.loader.Loader._instance_analysis')
    def test_check_module(self, mock_instance_analysis):
        # Setup
        mock_instance = MagicMock()
        mock_instance.get_input_params.return_value = [{'name': 'param1', 'default': 'value1'}]
        mock_instance.run.return_value = None
        mock_instance_analysis.return_value = ('test', mock_instance)
        # Call
        result = self.loader.check_module({'name': 'test'})
        # Assert
        self.assertTrue(result)

    def test_check_module_fail(self):
        # Call
        result = self.loader.check_module({'name': 'test'})
        # Assert
        self.assertFalse(result)

    def test_get_types(self):
        # Setup
        mock_handle = MagicMock()
        mock_handle.type.return_value = 'test_type'
        self.loader.analysis_handle = {'test': mock_handle}
        # Call
        result = self.loader.get_types()
        # Assert
        self.assertEqual(result, {'test': 'test_type'})

    @patch('neuronautics.analysis.loader.Loader._get_module')
    def test_load_and_execute_class(self, mock_get_module):
        # Setup
        mock_module = MagicMock()
        mock_get_module.return_value = mock_module
        mock_module.TestClass = MagicMock()
        mock_instance = mock_module.TestClass()
        mock_instance.get_input_params.return_value = 'test_input_params'
        mock_instance.execute.return_value = 'test_execute'
        # Call
        result = self.loader.load_and_execute_class('test_module', 'TestClass')
        # Assert
        self.assertEqual(result, ('test_input_params', 'test_execute'))

    def test_load_and_execute_class_exception(self):
        result, execution = self.loader.load_and_execute_class('test_module', 'TestClass')
        # Assert
        self.assertEqual(execution, "")
        self.assertTrue(result.startswith("Error"))

    @patch('neuronautics.analysis.loader.importlib.import_module')
    def test__get_module_import_module(self, mock_import_module):
        # Setup
        mock_module = MagicMock()
        mock_import_module.return_value = mock_module
        # Call
        result = self.loader._get_module('test_module', 'test_path')
        # Assert
        self.assertEqual(result, mock_module)

    @patch('neuronautics.analysis.loader.importlib.import_module', side_effect=Exception())
    @patch('neuronautics.analysis.loader.importlib.util.spec_from_file_location')
    @patch('neuronautics.analysis.loader.importlib.util.module_from_spec')
    def test__get_module(self, mock_module_from_spec, mock_spec_from_file_location, mock_import_module):
        # Setup
        mock_module = MagicMock()
        mock_import_module.return_value = mock_module
        mock_spec = MagicMock()
        mock_spec_from_file_location.return_value = mock_spec
        mock_module_from_spec.return_value = mock_module
        # Call
        result = self.loader._get_module('test_module', 'test_path')
        # Assert
        self.assertEqual(result, mock_module_from_spec.return_value)

    @patch('neuronautics.analysis.loader.Loader.check_module')
    @patch('neuronautics.analysis.loader.load_yaml')
    def test_save(self, mock_load_yaml, mock_check_module):
        # Setup
        Loader._dump = MagicMock(return_value=None)
        mock_check_module.return_value = True
        # Call
        self.loader.save({'name': 'test'})
        # Assert
        Loader._dump.assert_called_with(mock_load_yaml.return_value)

    @patch('neuronautics.analysis.loader.load_yaml')
    @patch('neuronautics.analysis.loader.yaml.dump')
    def test_delete(self, mock_dump, mock_load_yaml):
        # Setup
        mock_load_yaml.return_value = [{'name': 'test'}]
        # Call
        self.loader.delete('test')
        # Assert
        mock_dump.assert_called_once()

if __name__ == '__main__':
    unittest.main()



from ..analysis.loader import Loader
from ..analysis.type.abstract_analysis import AnalysisType
from ..analysis.analysis_creator import AnalysisCreatorUi
from ..utils.helpers import open_external_editor
from ..ui.code_error_ui import CodeErrorUi

# plotters
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas

# logger
from ..utils.logger import Logger

# pyqt
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QTreeWidgetItem

from ..utils.singleton import Singleton
from ..ui.helpers import clear_layout


class AnalysisUi(metaclass=Singleton):
    me = None

    def __init__(self, ui):
        self.ui = ui
        self.current_analysis = None
        self.current_analysis_params = None
        self.current_fig = None
        self.current_fig_metadata = None

    def load_analysis(self):
        tree = self.ui.analysisTree.invisibleRootItem()
        self.ui.loader.load()
        analysis = self.ui.loader.get_types()

        analysis_by_type = dict()
        for name, analysis_type in analysis.items():
            aux = analysis_by_type.get(analysis_type.value, [])
            aux.append(name)
            analysis_by_type[analysis_type.value] = aux

        for child_id in range(tree.childCount()):
            child = tree.child(child_id)
            for name in analysis_by_type.get(child.text(0), []):
                QTreeWidgetItem(child, [name])
            child.setExpanded(True)

    def analysis_selected(self, item, index):
        analysis_name = item.text(0)
        self.current_analysis = analysis_name
        loader = Loader()
        loader.load()
        handle = loader.analysis_handle.get(analysis_name)
        if handle:
            params = handle.get_input_params()
            layout = self.ui.inputAnalysisLyt
            clear_layout(layout)

            self.current_analysis_params = {}
            field_builder = dict(
                int=self._setup_int_field,
                float=self._setup_float_field,
                bool=self._setup_bool_field,
                list=self._setup_list_field
            )
            for param in params:
                label = QtWidgets.QLabel(param['name'])
                field = field_builder.get(param['type'])(param)
                self.current_analysis_params[param['name']] = field
                layout.addRow(label, field)

    @staticmethod
    def _setup_list_field(param):
        field = QtWidgets.QComboBox()
        field.addItems(param['values'])
        field.setCurrentText(param['default'])
        return field

    @staticmethod
    def _setup_bool_field(param):
        field = QtWidgets.QCheckBox()
        field.setChecked(param['default'])
        return field

    @staticmethod
    def _setup_float_field(param):
        field = QtWidgets.QDoubleSpinBox()
        field.setMinimum(param['min'])
        field.setMaximum(param['max'])
        field.setValue(param.get('default', param['min']))
        field.setSingleStep((param['max'] - param['min']) / 10)
        return field

    @staticmethod
    def _setup_int_field(param):
        field = QtWidgets.QSpinBox()
        field.setMinimum(param['min'])
        field.setMaximum(param['max'])
        field.setValue(param.get('default', param['min']))
        return field

    @staticmethod
    def _get_value(field):
        if isinstance(field, QtWidgets.QCheckBox):
            return field.isChecked()
        elif isinstance(field, QtWidgets.QComboBox):
            return field.currentText()
        return field.value()

    def run_analysis(self):
        current_analysis_params = {name: self._get_value(field) for name, field in self.current_analysis_params.items()}
        handle = Loader().analysis_handle.get(self.current_analysis)

        try:
            fig = handle.plot(spikes=self.ui.spike_handler.all_data(), **current_analysis_params)
            clear_layout(self.ui.outputAnalysisLyt)
            canvas = FigureCanvas(fig)
            self.ui.outputAnalysisLyt.addWidget(canvas)

            self.current_fig = fig
            self.current_fig_metadata = current_analysis_params
            self.ui.analysisExportBtn.setEnabled(True)
        except Exception as e:
            CodeErrorUi.show_exception()



    def new_analysis(self):
        analysis_ui = AnalysisCreatorUi()
        if analysis_ui.exec_():
            self.clear_analysis_tree()
            self.load_analysis()

    def clear_analysis_tree(self):
        tree = self.ui.analysisTree.invisibleRootItem()
        for i in reversed(range(tree.childCount())):
            for j in reversed(range(tree.child(i).childCount())):
                tree.child(i).takeChild(j)

    def delete_analysis(self):
        selected_analysis = self.ui.analysisTree.currentItem().text(0)
        self.ui.loader.delete(selected_analysis)
        self.clear_analysis_tree()
        self.load_analysis()

    def edit_analysis(self):
        selected_analysis = self.ui.analysisTree.currentItem().text(0)
        path = self.ui.loader.get_analysis_path(selected_analysis)
        if selected_analysis:
            open_external_editor(path)

    def export_analysis(self):
        if self.current_fig is not None:
            filename, filt = QtWidgets.QFileDialog.getSaveFileName(self.ui, "Save figure",
                                                                   options=QtWidgets.QFileDialog.DontUseNativeDialog,
                                                                   filter='Figure (*.png)')
            if filename and not filename.endswith('.png'):
                filename = filename + '.png'
            metadata = {f'PARAM: {k}': str(v) for k,v in self.current_fig_metadata.items()}
            self.current_fig.savefig(filename, format='png', transparent=True,
                                     backend='agg',
                                     metadata=metadata)



from PyQt5.QtChart import QLineSeries
from PyQt5.QtGui import QColor
import numpy as np


signal_color = {
    -3: (0, 0, 0, 0),      # invisible
    -2: (30, 144, 255),    # Blue       -- line selected
    -1: (200, 200, 200),   # Light Gray -- noise
    0: (50, 50, 50),       # Black
    1: (230, 85, 85),      # Soft Red
    2: (85, 230, 85),      # Soft Green
    3: (85, 85, 230),      # Soft Blue
    4: (230, 185, 85),     # Soft Orange
    5: (180, 85, 230),     # Soft Purple
    6: (230, 230, 85),     # Soft Yellow
    7: (85, 230, 230),     # Soft Cyan
    8: (230, 85, 230),     # Soft Magenta
    9: (150, 150, 150),    # Soft Gray
    10: (230, 160, 180),   # Soft Pink
    1000: (5, 71, 00),  # Blue
}


class SignalLine(QLineSeries):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.signal_id = 0
        self.index = 0
        self.visible = True
        self.y_range = None
        self.x_range = None

    @classmethod
    def color(cls, sid: int):
        return signal_color.get(sid, (0, 0, 0))

    def _change_color(self, cid: int):
        pen = self.pen()
        pen.setColor(QColor(*self.color(cid)))
        self.setPen(pen)

    def select(self):
        self._change_color(-2)

    def unselect(self):
        self._change_color(self.signal_id)

    def set_id(self, signal_id: int):
        self.signal_id = signal_id
        self._change_color(self.signal_id)

    def set_visible(self, visibility: bool):
        self.visible = visibility
        if self.visible:
            self._change_color(self.signal_id)
        else:
            self._change_color(-3)

    def set_index(self, index):
        self.index = index

    def has_points_inside_bounding_box(self, x_start, x_end, y_start, y_end):
        if not self.visible:
            return False

        x_start, x_end = min(x_start, x_end), max(x_start, x_end)
        y_start, y_end = min(y_start, y_end), max(y_start, y_end)

        # Check if any points in the series fall within the bounding box
        for point in self.pointsVector():
            x = point.x()
            y = point.y()
            if x_start <= x <= x_end and y_start <= y <= y_end:
                return True
        return False

    def append(self, *args, **kwargs):
        super().append(*args, **kwargs)
        ys = np.array([point.y() for point in self.pointsVector()])
        xs = np.array([point.x() for point in self.pointsVector()])
        self.y_range = (ys.min(), ys.max())
        self.x_range = (xs.min(), xs.max())



# NOTE: needs to create an app for the testing, but then other tests start failing. I will keep it commented for the future
#
# import unittest
# from PyQt5 import QtWidgets
# from unittest.mock import patch
# from neuronautics.analysis.analysis_creator import AnalysisCreatorUi
#
#
# class TestAnalysisCreatorUi(unittest.TestCase):
#     def setUp(self):
#         self.app = QtWidgets.QApplication([])
#         self.dialog = AnalysisCreatorUi()
#
#     def tearDown(self):
#         self.app.quit()
#
#     @patch('neuronautics.analysis.analysis_creator.open_external_editor')
#     @patch('neuronautics.analysis.analysis_creator.AnalysisCreator.create')
#     @patch('neuronautics.utils.logger.Logger')
#     def test_create_analysis(self, mock_logger, mock_create, mock_open_editor):
#         mock_create.return_value = '/path/to/analysis'
#         self.dialog.analysisName.setText('Test Analysis')
#
#         self.dialog.create_analysis()
#
#         mock_create.assert_called_once_with('Test Analysis', 'line')
#         mock_open_editor.assert_called_once_with('/path/to/analysis')
#
# if __name__ == '__main__':
#     unittest.main()



import importlib
import importlib.util
import yaml
from pathlib import Path
from ..utils.helpers import file_path, load_yaml, app_path
from ..utils.singleton import Singleton

from ..recordings.config import YML_ANALYSIS_BASE


BASE_ANALYTICS = YML_ANALYSIS_BASE
EXTRA_ANALYTICS = file_path('analysis/extra.yml')
# TODO: raised exceptions can be customised


class Loader(metaclass=Singleton):
    """
    A Singleton class that loads and manages analysis configurations.

    Attributes:
        analysis_handle (dict): A dictionary to hold the analysis handles.
        analysis_config (dict): A dictionary to hold the analysis configurations.
    """
    def __init__(self):
        """
        Initializes the Loader instance.
        """
        self.analysis_handle = None
        self.analysis_config = None

    def load(self):
        """
        Loads the analysis configuration from the base and extra analytics files.
        """
        base = load_yaml(BASE_ANALYTICS, list())
        analysis = load_yaml(EXTRA_ANALYTICS, list())
        for b in base:
            if not any([b.get('name') == an.get('name') for an in analysis]):
                analysis.append(b)
        analysis = sorted(analysis, key=lambda an: an.get('name'))

        self.analysis_handle = {name: handle for name, handle in [Loader._instance_analysis(an) for an in analysis]}
        self.analysis_config = {config.get('name'): config for config in analysis}

    def get_analysis_path(self, name: str) -> str:
        """
        Gets the path of the analysis.

        Args:
            name (str): The name of the analysis.

        Returns:
            str: The path of the analysis.
        """
        path = self.analysis_config[name].get('path')
        if path:
            return file_path(path)
        return None

    @classmethod
    def _instance_analysis(cls, analysis: dict) -> tuple:
        """
        Instantiates the analysis.

        Args:
            analysis (dict): The analysis configuration.

        Returns:
            tuple: A tuple containing the name of the analysis and the dynamic class instance.
        """

        try:
            module = cls._get_module(analysis.get('module'), analysis.get('path'))
            dynamic_class = getattr(module, analysis.get('class'))()
            return analysis.get('name'), dynamic_class
        except Exception as e:
            raise e

    @classmethod
    def check_module(cls, analysis: dict) -> bool:
        """
        Checks the module.

        Args:
            analysis (dict): The analysis configuration.

        Returns:
            bool: True if the module is valid, False otherwise.
        """
        try:
            name, instance = cls._instance_analysis(analysis)
            params_def = instance.get_input_params()
            params = {param['name']: param['default'] for param in params_def}
            instance.run(None, **params)
        except:
            return False
        return True

    def get_types(self) -> dict:
        """
        Gets the types of the analysis.

        Returns:
            dict: A dictionary mapping the name of the analysis to its type.
        """
        return {name: handle.type() for name, handle in self.analysis_handle.items()}

    @staticmethod
    def load_and_execute_class(module_name: str, class_name: str) -> tuple:
        """
        Loads and executes a class.

        Args:
            module_name (str): The name of the module.
            class_name (str): The name of the class.

        Returns:
            tuple: A tuple containing the result of the input parameters and the execution.
        """
        try:
            module = Loader._get_module(module_name)
            dynamic_class = getattr(module, class_name)
            instance = dynamic_class()

            input_params_result = instance.get_input_params()
            execute_result = instance.execute()

            return input_params_result, execute_result
        except Exception as e:
            return f"Error: {e}", ""

    @staticmethod
    def _get_module(module_name: str, path: str = None):
        """
        Gets the module.

        Args:
            module_name (str): The name of the module.
            path (str, optional): The path of the module. Defaults to None.

        Returns:
            module: The module.
        """
        try:
            module = importlib.import_module(module_name)
        except:
            spec = importlib.util.spec_from_file_location(module_name, file_path(path))
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
        return module

    @classmethod
    def save(cls, config: dict):
        """
        Saves the configuration.

        Args:
            config (dict): The configuration to save.
        """
        cls.delete(config['name'])
        analysis = load_yaml(EXTRA_ANALYTICS, list())
        if Loader.check_module(config):
            analysis.append(config)
            cls._dump(analysis)

    @classmethod
    def delete(cls, name: str):
        """
        Deletes the analysis.

        Args:
            name (str): The name of the analysis to delete.
        """
        analysis = load_yaml(EXTRA_ANALYTICS, list())
        analysis = [an for an in analysis if an.get('name').upper() != name.upper()]
        cls._dump(analysis)

    @classmethod
    def _dump(cls, analysis: list):
        """
        Dumps the analysis.

        Args:
            analysis (list): The analysis to dump.
        """
        with open(EXTRA_ANALYTICS, 'w') as stream:
            yaml.dump(analysis, stream)




import McsPy.McsData
import McsPy.McsCMOS

from ..recordings.abstract_raw import AbstractRaw
from ..recordings.config import MAX_WORKERS

import numpy as np
import pandas as pd
import concurrent.futures
import os

from ..utils.logger import Logger
from ..utils.helpers import parse_numeric_array
from ..utils.helpers import moving_stats, butter_lowpass_filter
from functools import lru_cache


TMP_FILE = '/tmp/spikes.csv'
TMP_PARTIAL_FILE = lambda fid: f'/tmp/spikes_{fid}.csv'


@lru_cache()
def fetch_raw(filename):
    return McsPy.McsData.RawData(filename)


def parallel_extract_spike(filename, ch_id, sigma, window_ms, pre_spike_ms, pos_spike_ms):
    spikes = McsRaw(filename).extract_spike(ch_id, sigma, window_ms, pre_spike_ms, pos_spike_ms)
    filename = TMP_PARTIAL_FILE(ch_id)
    spikes.to_csv(filename, index=False)
    return filename


def parallel_get_data(filename, ch_id, from_ix, to_ix):
    return ch_id, McsRaw(filename).get_channel_data(ch_id, from_ix, to_ix)


class McsRaw(AbstractRaw):
    logger = Logger.get_logger()

    def __init__(self, filename):
        self.filename = filename
        self.data = fetch_raw(filename)

    def __getstate__(self):
        return {'filename': self.filename}

    def __setstate__(self, state):
        self.filename = state['filename']
        self.data = fetch_raw(self.filename)

    @classmethod
    def _get_recording_info(cls, rec_id, rec) -> dict:
        recording_time = 0
        recording_unit = ''
        units = ['h', 'min', 's', 'ms']
        for unit in units:
            recording_time = rec.duration_time.m_as(unit)
            recording_unit = unit
            if recording_time > 1:
                break

        return {
            f'[{rec_id}] recording time': f'{recording_time:.2f} {recording_unit}',
        }

    def _get_recording(self, recording_id=None):
        recording_id = list(self.data.recordings.keys())[0] if recording_id is None else recording_id
        return self.data.recordings[recording_id]

    def _get_analog_stream(self, analog_stream_id=None):
        analog_stream_id = list(self._get_recording().analog_streams.keys())[0] if analog_stream_id is None \
                                                                                else analog_stream_id
        return self._get_recording().analog_streams[analog_stream_id]

    def get_analog_channels_ids(self):
        return list(self._get_analog_stream().channel_infos.keys())

    def get_info(self):
        data = self.data
        info = {
            'MEA Layout': data.mea_layout,
            'recording date': str(data.date),
            'comment': data.comment,
            'recordings ids': ', '.join([str(k) for k in data.recordings.keys()]),
        }
        for k in data.recordings.keys():
            info = {**info, **self._get_recording_info(k, data.recordings[k])}

        return info

    @classmethod
    def _to_microvolts(cls, data, units):
        factors = {'volt': 1_000_000}
        return data*factors[str(units)]

    def to_seconds(self, data, units):
        factors = {'microsecond': 1/1_000_000,
                   'millisecond': 1/1_000,
                   'second'     : 1,
                   'minute'     : 60,
                   'hour'       : 3600,
                   'tick'       : self._get_analog_stream().channel_infos[0].sampling_tick.m_as('millisecond')/1_000}
        return data*factors[str(units)]

    def to_ticks(self, data, units):
        data = self.to_seconds(data, units)
        tick = self._get_analog_stream().channel_infos[0].sampling_tick.m_as('second')
        return int(data / tick)

    def get_channel_data(self, channel_id, from_ix, to_ix):
        series, unit_v = self._get_analog_stream().get_channel_in_range(channel_id, from_ix, to_ix)
        time, unit_t = self._get_analog_stream().get_channel_sample_timestamps(0, from_ix, to_ix)
        series = self._to_microvolts(series, unit_v)
        time = self.to_seconds(time, unit_t)
        return time, series

    def get_all_data(self, from_ix, to_ix):
        channel_data = {}
        with concurrent.futures.ProcessPoolExecutor(max_workers=MAX_WORKERS) as executor:
            futures = [executor.submit(parallel_get_data,
                                       self.filename, ch_id, from_ix, to_ix)
                       for ch_id in self.get_analog_channels_ids()]
            for future in concurrent.futures.as_completed(futures):
                ch_id, spikes = future.result()
                channel_data[ch_id] = spikes
        return channel_data

    def extract_spike(self, channel_id, sigma, window_ms, pre_spike_ms=1, pos_spike_ms=2):
        try:
            window_ticks = self.to_ticks(window_ms, 'millisecond')
            pre_spike_ticks = self.to_ticks(pre_spike_ms, 'millisecond')
            pos_spike_ticks = self.to_ticks(pos_spike_ms, 'millisecond')
            channel_data, unit_v = self._get_analog_stream().get_channel_in_range(channel_id, 0, self._get_recording().duration)
            channel_data = self._to_microvolts(channel_data, unit_v)
            channel_data = channel_data - channel_data.mean()
            channel_data = butter_lowpass_filter(channel_data, 30, 10_000) # TODO: hardcoded values

            if window_ms == 0:
                ma = channel_data.mean()
                ms = channel_data.std()
                mask = np.abs(channel_data) > (ma + sigma*ms)
            else:
                ma, ms = moving_stats(channel_data, window_ticks)
                mask = np.abs(channel_data[-len(ma):]) > (ma + sigma*ms)
            indexes = np.where(mask)[0]

            aux_index = 0
            spikes = list()
            for index in indexes:
                if index > aux_index and (index < len(channel_data)-window_ticks):
                    rel_pos = window_ticks+index
                    shift = np.abs(channel_data[rel_pos:rel_pos+pre_spike_ticks]).argmax()
                    if shift == 0:
                        # shift 0 means there is only one value surpassing the threshold
                        # -> most likely noise
                        # -> or the continuation of a spike (spike + 1 ms)
                        continue
                    rel_pos = rel_pos + shift
                    shift_index = index + shift
                    aux_index = index + shift + pre_spike_ticks + 1
                    spike = channel_data[rel_pos-pre_spike_ticks:rel_pos+pos_spike_ticks]

                    if ms[index] > 0 and len(spike) == (pre_spike_ticks + pos_spike_ticks):
                        spikes.append((channel_id, rel_pos, self.to_seconds(rel_pos, 'tick') * 1_000
                                       , 0, ma[shift_index], ms[shift_index]
                                       , (abs(spike[pre_spike_ticks])-ma[shift_index]) / ms[shift_index], spike))

            spikes = pd.DataFrame(spikes, columns=['channel_id', 'tick', 'ts_ms', 'class', 'avg', 'std', 'sigma', 'spike'])

        except Exception as e:
            print(f'>>> Channel {channel_id} failed at {index}')
            spikes = pd.DataFrame(spikes, columns=['channel_id', 'tick', 'ts_ms', 'class', 'avg', 'std', 'sigma', 'spike'])
            raise e

        return spikes

    def extract_all_spikes(self, sigma, window_ms, pre_spike_ms=1, pos_spike_ms=2):
        spikes_files = []
        i = 1
        self.logger.log_process('extract_spike', 0, len(self.get_analog_channels_ids()))
        with concurrent.futures.ProcessPoolExecutor(max_workers=MAX_WORKERS) as executor:
            futures = [executor.submit(parallel_extract_spike,
                                       self.filename, ch_id, sigma, window_ms, pre_spike_ms, pos_spike_ms)
                       for ch_id in self.get_analog_channels_ids()]
            for future in concurrent.futures.as_completed(futures):
                spikes_files.append(future.result())
                self.logger.log_process('extract_spike', i, len(self.get_analog_channels_ids()))
                i += 1

        # Merge individual CSV files into a single CSV file
        with open(TMP_FILE, 'w') as outfile:
            for ix, filename in enumerate(spikes_files):
                with open(filename) as infile:
                    if ix==0:
                        lines = infile.readlines()
                    else:
                        lines = infile.readlines()[1:]
                    outfile.write("\n".join(lines))

        # Remove the individual CSV files
        for filename in spikes_files:
            os.remove(filename)

        # Read the merged CSV file into a DataFrame
        spikes = pd.read_csv(TMP_FILE)

        # Apply the convert_to_array function to the 'spike' column
        spikes['spike'] = spikes['spike'].apply(parse_numeric_array)
        return spikes




import subprocess
___version___="0.2.1.1.0"

import os

branch = (subprocess
          .run(["git", "rev-parse", "--abbrev-ref", "HEAD"],
               stdout=subprocess.PIPE, text=True)
          .stdout.strip())

branch = os.environ.get('BRANCH_NAME')
major, minor, patch, rc, dev = ___version___.split('.')
if branch in ('master', 'main'):
    __version__ = f'{major}.{minor}.{patch}'
elif branch in ('staging', ):
    __version__ = f'{major}.{minor}.{patch}rc{rc}'
else:
    __version__ = f'{major}.{minor}.{patch}dev{dev}'



import abc
from enum import Enum


class AnalysisType(Enum):
    LINE = 'line'
    GRAPH = 'graph'
    BAR = 'bar'
    IMAGE = 'image'
    OTHER = 'misc'


class AbstractAnalysis(abc.ABC):

    @abc.abstractmethod
    def get_input_params(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def type(self):
        # line, bar, graph, image
        raise NotImplementedError()

    @abc.abstractmethod
    def run(self, *args, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def plot(self, *args, **kwargs):
        raise NotImplementedError()

    def input_params_check(self, **kwargs):
        pass


from ..recordings.abstract_spike import AbstractSpike

import pandas as pd
from ..utils.helpers import parse_numeric_array
from functools import lru_cache

@lru_cache()
def fetch_spikes(filename):
    print('Fetching spikes')
    data = pd.read_csv(filename)
    data['channel_id'] = data.channel_id.astype(int)
    data['class'] = data['class'].astype(int)
    data['spike'] = data['spike'].apply(parse_numeric_array)
    return data


class NNSpike(AbstractSpike):

    def __init__(self, filename):
        self.filename = filename
        self.data = fetch_spikes(filename)

    def get_info(self):
        return {'N. channels': len(self.get_analog_channels_ids())}

    def get_analog_channels_ids(self):
        return sorted(list(self.data.channel_id.astype(int).drop_duplicates()))

    def get_channel_data(self, channel_id, from_ms, to_ms):
        channel_data = self.data[self.data.channel_id == channel_id]
        channel_data = channel_data[(from_ms<=channel_data.ts_ms) & (channel_data.ts_ms<to_ms)]
        return channel_data[['class', 'ts_ms', 'spike']]

    def get_all_data(self, from_ms, to_ms):
        channel_data = self.data
        channel_data = channel_data[(from_ms<=channel_data.ts_ms) & (channel_data.ts_ms<to_ms)]
        return channel_data[['channel_id', 'class', 'ts_ms', 'spike']]

    def set_labels(self, channel_id, labels):
        self.data.loc[self.data['channel_id'] == channel_id, 'class'] = labels

    def save(self):
        self.data.to_csv(self.filename, index=False)

    def to_ms(self, data, units):
        factors = { 'tick': 0.1 }
        return data*factors[str(units)]



from abc import ABC

from neuronautics.analysis.type.abstract_analysis import AbstractAnalysis, AnalysisType
from neuronautics.config.layout import Layout
from neuronautics.analysis.type.palette import Palette
import networkx as nx
import matplotlib.pyplot as plt


class GraphAnalysis(AbstractAnalysis, ABC):

    def type(self):
        return AnalysisType.GRAPH

    def plot(self, title='', *args, **kwargs):
        layout = Layout()
        layout.load()
        layout_mat = layout.current()
        electrode_positions = {int(id)-1: (j, i) for i, row in enumerate(layout_mat) for j, id in enumerate(row) if id}
        connectivity_matrix = self.run(*args, **kwargs)

        # Create a graph from the connectivity matrix
        G = nx.from_numpy_array(connectivity_matrix)

        # Create a figure and axis
        fig, ax = plt.subplots(figsize=(10, 8))

        # Draw the graph with fixed node positions
        nx.draw(G, pos=electrode_positions, with_labels=True, node_size=350, node_shape='h',
                node_color='white', edgecolors=Palette.node_edge_color, font_size=8, font_color=Palette.label_color,
                edge_color=Palette.link_color, connectionstyle='arc3,rad=0.4', arrows=True, ax=ax)

        # Set plot labels and title
        ax.set_aspect('equal')
        ax.set_title(title)

        return fig



import unittest
from unittest.mock import patch, MagicMock
from neuronautics.analysis.type.bar_analysis import BarAnalysis
from neuronautics.analysis.type.abstract_analysis import AnalysisType
from matplotlib.figure import Figure
import pandas as pd


class OneBarAnalysis(BarAnalysis):

    def run(self, *args, **kwargs):
        return pd.DataFrame({'A': [1,2], 'B': [3,4]})

    def get_input_params(self):
        pass


class TestBarAnalysis(unittest.TestCase):
    def test_type(self):
        analysis = OneBarAnalysis()
        self.assertEqual(analysis.type(), AnalysisType.BAR)

    @patch("matplotlib.pyplot.xticks")
    @patch("matplotlib.pyplot.legend")
    @patch("matplotlib.pyplot.title")
    @patch("matplotlib.pyplot.tight_layout")
    def test_plot_vertical(self, mock_xticks, mock_legend, mock_title, mock_tight_layout):
        # Act
        bar_analysis = OneBarAnalysis()
        result = bar_analysis.plot(value_units='', stacked=True, vertical=True, title='')

        # Assert
        self.assertIsInstance(result, Figure)
        mock_xticks.assert_called_once()
        mock_legend.assert_called_once()
        mock_title.assert_called_once()
        mock_tight_layout.assert_called_once()

    @patch("matplotlib.pyplot.yticks")
    @patch("matplotlib.pyplot.legend")
    @patch("matplotlib.pyplot.title")
    @patch("matplotlib.pyplot.tight_layout")
    def test_plot_horizontal(self, mock_yticks, mock_legend, mock_title, mock_tight_layout):
        # Act
        bar_analysis = OneBarAnalysis()
        result = bar_analysis.plot(value_units='', stacked=True, vertical=False, title='')

        # Assert
        self.assertIsInstance(result, Figure)
        mock_yticks.assert_called_once()
        mock_legend.assert_called_once()
        mock_title.assert_called_once()
        mock_tight_layout.assert_called_once()

    @patch("matplotlib.pyplot.yticks")
    @patch("matplotlib.pyplot.legend")
    @patch("matplotlib.pyplot.title")
    @patch("matplotlib.pyplot.tight_layout")
    def test_plot_no_stacked(self, mock_yticks, mock_legend, mock_title, mock_tight_layout):
        # Act
        bar_analysis = OneBarAnalysis()
        result = bar_analysis.plot(value_units='', stacked=False, vertical=False, title='')

        # Assert
        self.assertIsInstance(result, Figure)
        mock_yticks.assert_called_once()
        mock_legend.assert_called_once()
        mock_title.assert_called_once()
        mock_tight_layout.assert_called_once()


if __name__ == '__main__':
    unittest.main()



import unittest
from neuronautics.analysis.type.misc_analysis import MiscAnalysis
from neuronautics.analysis.type.abstract_analysis import AnalysisType


class OneMiscAnalysis(MiscAnalysis):

    def run(self, *args, **kwargs):
        pass

    def get_input_params(self):
        pass

    def plot(self, *args, **kwargs):
        pass


class TestBarAnalysis(unittest.TestCase):
    def test_type(self):
        analysis = OneMiscAnalysis()
        self.assertEqual(analysis.type(), AnalysisType.OTHER)



if __name__ == '__main__':
    unittest.main()
